#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os, sys, posix

from six import iteritems

from reclass import get_storage, output, get_path_mangler
from reclass.core import Core
from reclass.errors import ReclassException
from reclass.config import find_and_read_configfile, get_options
from reclass.constants import MODE_NODEINFO
from reclass.defaults import *
from reclass.settings import Settings
from reclass.version import *

def ext_pillar(minion_id, pillar,
               storage_type=OPT_STORAGE_TYPE,
               inventory_base_uri=OPT_INVENTORY_BASE_URI,
               nodes_uri=OPT_NODES_URI,
               classes_uri=OPT_CLASSES_URI,
               class_mappings=None,
               propagate_pillar_data_to_reclass=False,
               compose_node_name=OPT_COMPOSE_NODE_NAME,
               **kwargs):

    path_mangler = get_path_mangler(storage_type)
    nodes_uri, classes_uri = path_mangler(inventory_base_uri, nodes_uri, classes_uri)
    storage = get_storage(storage_type, nodes_uri, classes_uri, compose_node_name)
    input_data = None
    if propagate_pillar_data_to_reclass:
        input_data = pillar
    settings = Settings(kwargs)
    reclass = Core(storage, class_mappings, settings, input_data=input_data)

    data = reclass.nodeinfo(minion_id)
    params = data.get('parameters', {})
    params['__reclass__'] = {}
    params['__reclass__']['nodename'] = minion_id
    params['__reclass__']['applications'] = data['applications']
    params['__reclass__']['classes'] = data['classes']
    params['__reclass__']['environment'] = data['environment']
    return params


def top(minion_id, storage_type=OPT_STORAGE_TYPE,
        inventory_base_uri=OPT_INVENTORY_BASE_URI, nodes_uri=OPT_NODES_URI,
        classes_uri=OPT_CLASSES_URI, class_mappings=None, compose_node_name=OPT_COMPOSE_NODE_NAME,
        **kwargs):

    path_mangler = get_path_mangler(storage_type)
    nodes_uri, classes_uri = path_mangler(inventory_base_uri, nodes_uri, classes_uri)
    storage = get_storage(storage_type, nodes_uri, classes_uri, compose_node_name)
    settings = Settings(kwargs)
    reclass = Core(storage, class_mappings, settings, input_data=None)

    # if the minion_id is not None, then return just the applications for the
    # specific minion, otherwise return the entire top data (which we need for
    # CLI invocations of the adapter):
    if minion_id is not None:
        data = reclass.nodeinfo(minion_id)
        applications = data.get('applications', [])
        env = data['environment']
        return {env: applications}

    else:
        data = reclass.inventory()
        nodes = {}
        for (node_id, node_data) in iteritems(data['nodes']):
            env = node_data['environment']
            if env not in nodes:
                nodes[env] = {}
            nodes[env][node_id] = node_data['applications']

        return nodes


def cli():
    try:
        inventory_dir = os.path.abspath(os.path.dirname(sys.argv[0]))
        defaults = {'pretty_print' : True,
                    'no_refs' : False,
                    'output' : 'yaml',
                    'inventory_base_uri': inventory_dir
                   }
        defaults.update(find_and_read_configfile())
        options = get_options(RECLASS_NAME, VERSION, DESCRIPTION,
                              inventory_shortopt='-t',
                              inventory_longopt='--top',
                              inventory_help='output the state tops (inventory)',
                              nodeinfo_shortopt='-p',
                              nodeinfo_longopt='--pillar',
                              nodeinfo_dest='nodename',
                              nodeinfo_help='output pillar data for a specific node',
                              defaults=defaults)
        class_mappings = defaults.get('class_mappings')
        defaults.update(vars(options))
        defaults.pop("storage_type", None)
        defaults.pop("inventory_base_uri", None)
        defaults.pop("nodes_uri", None)
        defaults.pop("classes_uri", None)
        defaults.pop("class_mappings", None)

        if options.mode == MODE_NODEINFO:
            data = ext_pillar(options.nodename, {},
                              storage_type=options.storage_type,
                              inventory_base_uri=options.inventory_base_uri,
                              nodes_uri=options.nodes_uri,
                              classes_uri=options.classes_uri,
                              class_mappings=class_mappings,
                              **defaults)
        else:
            data = top(minion_id=None,
                       storage_type=options.storage_type,
                       inventory_base_uri=options.inventory_base_uri,
                       nodes_uri=options.nodes_uri,
                       classes_uri=options.classes_uri,
                       class_mappings=class_mappings,
                       **defaults)

        print(output(data, options.output, options.pretty_print, options.no_refs))

    except ReclassException as e:
        e.exit_with_message(sys.stderr)

    sys.exit(posix.EX_OK)

if __name__ == '__main__':
    cli()



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

RECLASS_NAME = 'reclass'
DESCRIPTION = ('merge data by recursive descent down an ancestry hierarchy '
               '(forked extended version)')
VERSION = '1.6.0'
AUTHOR = 'martin f. krafft / Andrew Pickford / salt-formulas community'
AUTHOR_EMAIL = 'salt-formulas@freelists.org'
MAINTAINER = 'salt-formulas community'
MAINTAINER_EMAIL = 'salt-formulas@freelists.org'
COPYRIGHT = ('Copyright © 2007–14 martin f. krafft, extensions © 2017 Andrew'
             ' Pickford, extensions © salt-formulas community')
LICENCE = 'Artistic Licence 2.0'
URL = 'https://github.com/salt-formulas/reclass'



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"jinja2 tests"

import base64
import unittest
import tempfile
import time
from kapitan.utils import render_jinja2_file
from kapitan.resources import inventory
from kapitan.inputs.jinja2_filters import base64_encode
from kapitan.refs.base import RefController, Revealer
from kapitan.refs.base64 import Base64Ref
from kapitan import cached
from collections import namedtuple


class Jinja2FiltersTest(unittest.TestCase):
    def test_sha256(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|sha256 }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "this and that"}
            output = "e863c1ac42619a2b429a08775a6acd89ff4c2c6b8dae12e3461a5fa63b2f92f5"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_base64_encode(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|b64encode }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "this and that"}
            output = "dGhpcyBhbmQgdGhhdA=="
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_base64_decode(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|b64decode }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "dGhpcyBhbmQgdGhhdA=="}
            output = "this and that"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_toml(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|toml }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": {"foo": ["this", "that"]}}
            output = 'foo = [ "this", "that",]\n'
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_yaml(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|yaml }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": ["this", "that"]}
            output = "- this\n- that\n"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_fileglob(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|fileglob }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "./tests/*jinja2.py"}
            output = "['./tests/test_jinja2.py']"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_bool(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|bool }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "yes"}
            output = "True"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_to_datetime(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|to_datetime }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "2019-03-07 13:37:00"}
            output = "2019-03-07 13:37:00"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_strftime(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|strftime }}".encode("UTF-8"))
            f.seek(0)
            format = "%a, %d %b %Y %H:%M"
            context = {"text": format}
            output = time.strftime(format)
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_regex_replace(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|regex_replace(pattern='world', replacement='kapitan') }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "hello world"}
            output = "hello kapitan"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_regex_escape(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|regex_escape }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "+s[a-z].*"}
            output = "\\+s\\[a\\-z\\]\\.\\*"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_regex_search(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|regex_search('world.*') }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "hello world"}
            output = "world"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_regex_findall(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|regex_findall('world.*') }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "hello world"}
            output = "['world']"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_ternary(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|ternary('yes', 'no') }}".encode("UTF-8"))
            f.seek(0)
            context = {"text": "kapitan == kapitan"}
            output = "yes"
            self.assertEqual(render_jinja2_file(f.name, context), output)

    def test_shuffle(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ text|shuffle }}".encode("UTF-8"))
            f.seek(0)
            array = [1, 2, 3, 4, 5]
            context = {"text": array}
            self.assertNotEqual(render_jinja2_file(f.name, context), array)

    def test_reveal_maybe_b64encode_tag(self):
        """
        creates ?{base64:some_value} and runs reveal_maybe|b64encode jinja2 filters
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ my_ref_tag_var|reveal_maybe|b64encode }}".encode("UTF-8"))
            f.seek(0)

            # new argparse namespace with --reveal and --refs-path values
            namespace = namedtuple("Namespace", [])
            namespace.reveal = True
            namespace.refs_path = tempfile.mkdtemp()

            # reveal_maybe uses cached, so inject namespace
            cached.args["compile"] = namespace
            cached.ref_controller_obj = RefController(cached.args["compile"].refs_path)
            cached.revealer_obj = Revealer(cached.ref_controller_obj)

            ref_tag = "?{base64:some_value}"
            ref_value = b"sitar_rock!"
            cached.ref_controller_obj[ref_tag] = Base64Ref(ref_value)
            context = {"my_ref_tag_var": ref_tag}
            ref_value_b64 = base64.b64encode(ref_value).decode()
            self.assertEqual(render_jinja2_file(f.name, context), ref_value_b64)

    def test_reveal_maybe_tag_no_reveal_flag(self):
        """
        creates ?{base64:some_value} and runs reveal_maybe jinja2 filters without --reveal flag
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ my_ref_tag_var|reveal_maybe }}".encode("UTF-8"))
            f.seek(0)

            # new argparse namespace with --reveal and --refs-path values
            namespace = namedtuple("Namespace", [])
            namespace.reveal = False
            namespace.refs_path = tempfile.mkdtemp()

            # reveal_maybe uses cached, so inject namespace
            cached.args["compile"] = namespace
            cached.ref_controller_obj = RefController(cached.args["compile"].refs_path)
            cached.revealer_obj = Revealer(cached.ref_controller_obj)

            ref_tag = "?{base64:some_value}"
            ref_value = b"sitar_rock!"
            cached.ref_controller_obj[ref_tag] = Base64Ref(ref_value)
            context = {"my_ref_tag_var": ref_tag}
            self.assertEqual(render_jinja2_file(f.name, context), "?{base64:some_value}")

    def test_reveal_maybe_no_tag(self):
        """
        runs reveal_maybe jinja2 filter on data without ref tags
        """
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ my_var|reveal_maybe }}".encode("UTF-8"))
            f.seek(0)

            # new argparse namespace with --reveal and --refs-path values
            namespace = namedtuple("Namespace", [])
            namespace.reveal = True
            namespace.refs_path = tempfile.mkdtemp()

            # reveal_maybe uses cached, so inject namespace
            cached.args["compile"] = namespace
            cached.ref_controller_obj = RefController(cached.args["compile"].refs_path)
            cached.revealer_obj = Revealer(cached.ref_controller_obj)

            var_value = "heavy_rock!"
            context = {"my_var": var_value}
            self.assertEqual(render_jinja2_file(f.name, context), var_value)


class Jinja2ContextVars(unittest.TestCase):
    def test_inventory_context(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ inventory.parameters.cluster.name }}".encode("UTF-8"))
            cluster_name = "minikube"
            target_name = "minikube-es"
            inv = inventory(["examples/kubernetes"], target_name)
            context = {"inventory": inv}
            f.seek(0)
            self.assertEqual(render_jinja2_file(f.name, context), cluster_name)

    def test_inventory_global_context(self):
        with tempfile.NamedTemporaryFile() as f:
            target_name = "minikube-es"
            f.write(
                '{{ inventory_global["%s"].parameters.cluster.name }}'.encode("UTF-8")
                % target_name.encode("UTF-8")
            )
            cluster_name = "minikube"
            inv_global = inventory(["examples/kubernetes"], None)
            context = {"inventory_global": inv_global}
            f.seek(0)
            self.assertEqual(render_jinja2_file(f.name, context), cluster_name)


class Jinja2ExternalFilterTest(unittest.TestCase):
    def test_custom_filter_jinja2(self):
        with tempfile.NamedTemporaryFile() as f:
            f.write("{{ inventory.parameters.cluster.name | custom_jinja2_filter }}".encode("UTF-8"))
            cluster_name = "minikube"
            target_name = "minikube-es"
            inv = inventory(["examples/kubernetes"], target_name)
            context = {"inventory": inv}
            f.seek(0)
            actual_output = render_jinja2_file(
                f.name, context, "./examples/kubernetes/lib/custom_jinja2_filter.py"
            )
            expected_output = base64_encode(cluster_name)
            self.assertEqual(actual_output, expected_output)



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys, os, posix

from reclass import get_storage, output
from reclass.core import Core
from reclass.settings import Settings
from reclass.config import find_and_read_configfile, get_options
from reclass.defaults import *
from reclass.errors import ReclassException
from reclass.constants import MODE_NODEINFO
from reclass.version import *

def main():
    try:
        defaults = {'no_refs' : OPT_NO_REFS,
                    'pretty_print' : OPT_PRETTY_PRINT,
                    'output' : OPT_OUTPUT
                   }
        defaults.update(find_and_read_configfile())

        options = get_options(RECLASS_NAME, VERSION, DESCRIPTION, defaults=defaults)
        storage = get_storage(options.storage_type,
                              options.nodes_uri,
                              options.classes_uri,
                              options.compose_node_name)
        class_mappings = defaults.get('class_mappings')
        defaults.update(vars(options))
        settings = Settings(defaults)
        reclass = Core(storage, class_mappings, settings)

        if options.mode == MODE_NODEINFO:
            data = reclass.nodeinfo(options.nodename)
        else:
            data = reclass.inventory()

        print(output(data, options.output, options.pretty_print, options.no_refs))

    except ReclassException as e:
        e.exit_with_message(sys.stderr)

    sys.exit(posix.EX_OK)

if __name__ == '__main__':
    main()



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"hashicorp vault kv secrets module"

import base64
import logging
import os
from binascii import Error as b_error
from sys import exit

from kapitan import cached
from kapitan.errors import KapitanError
from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend

import hvac
from hvac.exceptions import Forbidden, InvalidPath

logger = logging.getLogger(__name__)


class VaultError(KapitanError):
    """Generic vault errors"""

    pass


def get_env(parameter):
    """
    The following variables need to be exported to the environment or defined in inventory.
        * VAULT_ADDR: url for vault
        * VAULT_SKIP_VERIFY=true: if set, do not verify presented TLS certificate before communicating with Vault server.
        * VAULT_CLIENT_KEY: path to an unencrypted PEM-encoded private key matching the client certificate
        * VAULT_CLIENT_CERT: path to a PEM-encoded client certificate for TLS authentication to the Vault server
        * VAULT_CACERT: path to a PEM-encoded CA cert file to use to verify the Vault server TLS certificate
        * VAULT_CAPATH: path to a directory of PEM-encoded CA cert files to verify the Vault server TLS certificate
        * VAULT_NAMESPACE: specify the Vault Namespace, if you have one

    Following keys are used to creates a new hvac client instance.
        :param url: Base URL for the Vault instance being addressed.
        :type url: str
        :param cert: Certificates for use in requests sent to the Vault instance. This should be a tuple with the
            certificate and then key.
        :type cert: tuple
        :param verify: Either a boolean to indicate whether TLS verification should be performed when sending requests to Vault,
            or a string pointing at the CA bundle to use for verification.
            See http://docs.python-requests.org/en/master/user/advanced/#ssl-cert-verification.
        :type verify: Union[bool,str]
        :param namespace: Optional Vault Namespace.
        :type namespace: str
    """
    client_parameters = {}
    client_parameters["url"] = parameter.get("VAULT_ADDR", os.getenv("VAULT_ADDR", default=""))
    client_parameters["namespace"] = parameter.get(
        "VAULT_NAMESPACE", os.getenv("VAULT_NAMESPACE", default="")
    )
    # VERIFY VAULT SERVER TLS CERTIFICATE
    skip_verify = parameter.get("VAULT_SKIP_VERIFY", os.getenv("VAULT_SKIP_VERIFY", default=""))

    if skip_verify.lower() == "false":
        cert = parameter.get("VAULT_CACERT", os.getenv("VAULT_CACERT", default=""))
        if not cert:
            cert_path = parameter.get("VAULT_CAPATH", os.getenv("VAULT_CAPATH", default=""))
            if not cert_path:
                raise Exception("Neither VAULT_CACERT nor VAULT_CAPATH specified")
            client_parameters["verify"] = cert_path
        else:
            client_parameters["verify"] = cert
    else:
        client_parameters["verify"] = False

    # CLIENT CERTIFICATE FOR TLS AUTHENTICATION
    client_key = parameter.get("VAULT_CLIENT_KEY", os.getenv("VAULT_CLIENT_KEY", default=""))
    client_cert = parameter.get("VAULT_CLIENT_CERT", os.getenv("VAULT_CLIENT_CERT", default=""))
    if client_key != "" and client_cert != "":
        client_parameters["cert"] = (client_cert, client_key)
    return client_parameters


def vault_obj(vault_parameters):
    """
    vault_parameters: necessary parameters to authenticate & get value from vault, provided by inventory
    e.g.:
        auth: userpass
        VAULT_ADDR: http://127.0.0.1:8200
        VAULT_SKIP_VERIFY: false
    Authenticate client to server and return client object
    """
    env = get_env(vault_parameters)

    client = hvac.Client(**{k: v for k, v in env.items() if k != "auth"})

    auth_type = vault_parameters["auth"]
    # GET TOKEN EITHER FROM ENVIRONMENT OF FILE
    if auth_type in ["token", "github"]:
        env["token"] = os.getenv("VAULT_TOKEN")
        if not env["token"]:
            try:
                token_file = os.path.join(os.path.expanduser("~"), ".vault-token")
                with open(token_file, "r") as f:
                    env["token"] = f.read()
                if env["token"] == "":
                    raise VaultError("{file} is empty".format(file=token_file))
            except IOError:
                raise VaultError("Cannot read file {file}".format(file=token_file))
    # DIFFERENT LOGIN METHOD BASED ON AUTHENTICATION TYPE
    if auth_type == "token":
        client.token = env["token"]
    elif auth_type == "ldap":
        client.auth.ldap.login(username=os.getenv("VAULT_USERNAME"), password=os.getenv("VAULT_PASSWORD"))
    elif auth_type == "userpass":
        client.auth_userpass(username=os.getenv("VAULT_USERNAME"), password=os.getenv("VAULT_PASSWORD"))
    elif auth_type == "approle":
        client.auth_approle(os.getenv("VAULT_ROLE_ID"), secret_id=os.getenv("VAULT_SECRET_ID"))
    elif auth_type == "github":
        client.auth.github.login(token=env["token"])
    else:
        raise "Authentication type '{auth}' not supported".format(auth=auth_type)

    if client.is_authenticated():
        return client
    else:
        raise VaultError("Vault Authentication Error, Environment Variables defined?")


class VaultSecret(Base64Ref):
    """
    Hashicorp Vault support for KV Secret Engine
    """

    def __init__(self, data, vault_params, **kwargs):
        """
        Set vault parameter and encoding of data
        """
        self.data = data
        self.vault_params = vault_params
        super().__init__(self.data, **kwargs)
        self.type_name = "vaultkv"

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new VaultSecret from data and ref_params: target_name
        parameters will be grabbed from the inventory via target_name
        """
        try:
            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv["nodes"].get(target_name, None)
            if target_inv is None:
                raise ValueError("target_inv not set")

            ref_params.kwargs["vault_params"] = target_inv["parameters"]["kapitan"]["secrets"]["vaultkv"]
            return cls(data, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create VaultSecret: vaultkv parameters missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        Returns decrypted data
        """
        # can't use super().reveal() as we want bytes
        try:
            self.data = base64.b64decode(self.data, validate=True)
        except b_error:
            exit("non-alphabet characters in the data")

        return self._decrypt()

    def _decrypt(self):
        """
        Authenticate with Vault server & returns value of the key from secret

        :returns: secret in plain text
        """
        try:
            client = vault_obj(self.vault_params)
            # token will comprise of two parts path_in_vault:key
            data = self.data.decode("utf-8").rstrip().split(":")
            return_data = ""
            if self.vault_params.get("engine") == "kv":
                response = client.secrets.kv.v1.read_secret(
                    path=data[0], mount_point=self.vault_params.get("mount", "secret")
                )
                return_data = response["data"][data[1]]
            else:
                response = client.secrets.kv.v2.read_secret_version(
                    path=data[0], mount_point=self.vault_params.get("mount", "secret")
                )
                return_data = response["data"]["data"][data[1]]
            client.adapter.close()
        except Forbidden:
            raise VaultError(
                "Permission Denied. "
                + "make sure the token is authorised to access {path} on Vault".format(path=data[0])
            )
        except InvalidPath:
            raise VaultError("{path} does not exist on Vault secret".format(path=data[0]))

        if return_data == "":
            raise VaultError("'{key}' doesn't exist on '{path}'".format(key=data[1], path=data[0]))
        return return_data

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {
            "data": self.data,
            "encoding": self.encoding,
            "type": self.type_name,
            "vault_params": self.vault_params,
        }


class VaultBackend(Base64RefBackend):
    def __init__(self, path, ref_type=VaultSecret, **ref_kwargs):
        "init VaultBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = "vaultkv"



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import logging
import os

from kapitan.inputs.base import CompiledFile, InputType
from kapitan.resources import inventory
from kapitan.utils import render_jinja2

logger = logging.getLogger(__name__)


class Jinja2(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("jinja2", compile_path, search_paths, ref_controller)

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write items in path as jinja2 rendered files to compile_path.
        path can be either a file or directory.
        kwargs:
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
        """
        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)

        # set ext_vars and inventory for jinja2 context
        context = ext_vars.copy()
        context["inventory"] = inventory(self.search_paths, target_name)
        context["inventory_global"] = inventory(self.search_paths, None)
        jinja2_filters = kwargs.get("jinja2_filters")

        for item_key, item_value in render_jinja2(
            file_path, context, jinja2_filters=jinja2_filters, search_paths=self.search_paths
        ).items():
            full_item_path = os.path.join(compile_path, item_key)

            with CompiledFile(
                full_item_path, self.ref_controller, mode="w", reveal=reveal, target_name=target_name
            ) as fp:
                fp.write(item_value["content"])
                mode = item_value["mode"]
                os.chmod(full_item_path, mode)
                logger.debug("Wrote %s with mode %.4o", full_item_path, mode)

    def default_output_type(self):
        # no output_type options for jinja2
        return None



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.storage.common import NameMangler

class NodeStorageBase(object):

    def __init__(self, name):
        self._name = name

    name = property(lambda self: self._name)

    def get_node(self, name, settings):
        msg = "Storage class '{0}' does not implement node entity retrieval."
        raise NotImplementedError(msg.format(self.name))

    def get_class(self, name, environment, settings):
        msg = "Storage class '{0}' does not implement class entity retrieval."
        raise NotImplementedError(msg.format(self.name))

    def enumerate_nodes(self):
        msg = "Storage class '{0}' does not implement node enumeration."
        raise NotImplementedError(msg.format(self.name))

    def path_mangler(self):
        msg = "Storage class '{0}' does not implement path_mangler."
        raise NotImplementedError(msg.format(self.name))


class ExternalNodeStorageBase(NodeStorageBase):

    def __init__(self, name, compose_node_name):
        super(ExternalNodeStorageBase, self).__init__(name)
        self.class_name_mangler = NameMangler.classes
        if compose_node_name:
            self.node_name_mangler = NameMangler.composed_nodes
        else:
            self.node_name_mangler = NameMangler.nodes



# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import hashlib
import logging
import os
import re
import tarfile
import tempfile
from collections import defaultdict
from distutils.dir_util import copy_tree
from functools import partial
from shutil import copyfile
from zipfile import ZipFile
import platform

from kapitan.errors import GitSubdirNotFoundError, HelmBindingUnavailableError
from kapitan.utils import make_request

from git import Repo

logger = logging.getLogger(__name__)

try:
    from kapitan.dependency_manager.helm.helm_fetch_binding import ffi
except ImportError as ie:
    logger.debug("Error importing ffi from kapitan.dependency_manager.helm.helm_fetch_binding: {}".format(ie))
    pass  # make this feature optional

# this is used just to change the root directory of save_dir in testing
DEPENDENCY_OUTPUT_CONFIG = {}


def fetch_dependencies(target_objs, pool):
    """
    parses through the dependencies parameters in target_objs and fetches them
    all dependencies are first fetched into tmp dir, after which they are copied to their respective output_path.
    """
    temp_dir = tempfile.mkdtemp()
    # there could be multiple dependency items per source_uri due to reclass inheritance or
    # other user requirements. So create a mapping from source_uri to a set of dependencies with
    # that source_uri
    git_deps = defaultdict(list)
    http_deps = defaultdict(list)
    helm_deps = defaultdict(list)

    # this dict is to make sure no duplicated output_paths exist per source
    deps_output_paths = defaultdict(set)
    for target_obj in target_objs:
        try:
            dependencies = target_obj["dependencies"]
            for item in dependencies:
                dependency_type = item["type"]
                source_uri = item["source"]
                if "root_dir" in DEPENDENCY_OUTPUT_CONFIG:
                    # should only be used in test environment
                    item["output_path"] = os.path.join(
                        DEPENDENCY_OUTPUT_CONFIG["root_dir"], item["output_path"]
                    )
                output_path = item["output_path"]

                if output_path in deps_output_paths[source_uri]:
                    # if the output_path is duplicated for the same source_uri
                    continue
                else:
                    deps_output_paths[source_uri].add(output_path)

                if dependency_type == "git":
                    git_deps[source_uri].append(item)
                elif dependency_type in ("http", "https"):
                    http_deps[source_uri].append(item)
                elif dependency_type == "helm":
                    version = item.get("version", "")
                    if version == "":
                        version = "latest"
                    helm_deps[item["chart_name"] + "-" + version].append(item)

        except KeyError:
            continue

    git_worker = partial(fetch_git_dependency, save_dir=temp_dir)
    http_worker = partial(fetch_http_dependency, save_dir=temp_dir)
    helm_worker = partial(fetch_helm_chart, save_dir=temp_dir)
    [p.get() for p in pool.imap_unordered(http_worker, http_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(git_worker, git_deps.items()) if p]
    [p.get() for p in pool.imap_unordered(helm_worker, helm_deps.items()) if p]


def fetch_git_dependency(dep_mapping, save_dir):
    """
    fetches a git repository at source into save_dir, and copy the repository into output_path
    ref is used to checkout if exists
    only subdir is copied into output_path if specified
    """
    source, deps = dep_mapping
    fetch_git_source(source, save_dir)
    for dep in deps:
        repo_path = os.path.join(save_dir, os.path.basename(source))
        repo = Repo(repo_path)
        output_path = dep["output_path"]
        copy_src_path = repo_path
        if "ref" in dep:
            ref = dep["ref"]
            repo.git.checkout(ref)
        else:
            repo = Repo(repo_path)
            repo.git.checkout("master")  # default ref

        if "subdir" in dep:
            sub_dir = dep["subdir"]
            full_subdir = os.path.join(repo_path, sub_dir)
            if os.path.isdir(full_subdir):
                copy_src_path = full_subdir
            else:
                raise GitSubdirNotFoundError(
                    "Dependency {} : subdir {} not found in repo".format(source, sub_dir)
                )

        if not os.path.exists(os.path.abspath(output_path)):
            copy_tree(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_git_source(source, save_dir):
    """clones a git repository at source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    Repo.clone_from(source, os.path.join(save_dir, os.path.basename(source)))
    logger.info("Dependency {} : successfully fetched".format(source))


def fetch_http_dependency(dep_mapping, save_dir):
    """
    fetches a http[s] file at source and saves into save_dir, after which it is copied into
    the output_path stored in dep_mapping
    """
    source, deps = dep_mapping
    content_type = fetch_http_source(source, save_dir)
    path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
    copy_src_path = os.path.join(save_dir, path_hash + os.path.basename(source))
    for dep in deps:
        output_path = dep["output_path"]
        if dep.get("unpack", False):
            # ensure that the directory we are extracting to exists
            os.makedirs(output_path, exist_ok=True)
            is_unpacked = False
            if content_type == "application/x-tar":
                tar = tarfile.open(copy_src_path)
                tar.extractall(path=output_path)
                tar.close()
                is_unpacked = True
            elif content_type == "application/zip":
                zfile = ZipFile(copy_src_path)
                zfile.extractall(output_path)
                zfile.close()
                is_unpacked = True
            elif content_type in ["application/octet-stream", "application/x-gzip"]:
                if re.search(r"(\.tar\.gz|\.tgz)$", source):
                    tar = tarfile.open(copy_src_path)
                    tar.extractall(path=output_path)
                    tar.close()
                    is_unpacked = True
            if is_unpacked:
                logger.info("Dependency {} : extracted to {}".format(source, output_path))
            else:
                logger.info(
                    "Dependency {} : Content-Type {} is not supported for unpack. Ignoring save".format(
                        source, content_type
                    )
                )
        else:
            # we are downloading a single file
            parent_dir = os.path.dirname(output_path)
            if parent_dir != "":
                os.makedirs(parent_dir, exist_ok=True)
            copyfile(copy_src_path, output_path)
            logger.info("Dependency {} : saved to {}".format(source, output_path))


def fetch_http_source(source, save_dir):
    """downloads a http[s] file from source and saves into save_dir"""
    logger.info("Dependency {} : fetching now".format(source))
    content, content_type = make_request(source)
    logger.info("Dependency {} : successfully fetched".format(source))
    if content is not None:
        basename = os.path.basename(source)
        # to avoid collisions between basename(source)
        path_hash = hashlib.sha256(os.path.dirname(source).encode()).hexdigest()[:8]
        full_save_path = os.path.join(save_dir, path_hash + basename)
        with open(full_save_path, "wb") as f:
            f.write(content)
        return content_type
    else:
        logger.warning("Dependency {} : failed to fetch".format(source))
        return None


def fetch_helm_chart(dep_mapping, save_dir):
    """
    downloads a helm chart and its subcharts from source then untars and moves it to save_dir
    """
    lib = initialise_helm_fetch_binding()
    unique_chart, deps = dep_mapping
    for dep in deps:
        chart_name = dep["chart_name"]
        version = dep.get("version", "")
        repo_url = dep["source"]
        destination_dir = dep["output_path"]

        if version == "":
            logger.info(
                "Dependency helm chart {} being fetch with using latest version available".format(chart_name)
            )

        logger.info("Dependency helm chart {} and version {}: fetching now".format(chart_name, version))
        c_chart_name = ffi.new("char[]", chart_name.encode("ascii"))
        c_version = ffi.new("char[]", version.encode("ascii"))
        c_repo_url = ffi.new("char[]", repo_url.encode("ascii"))
        c_destination_dir = ffi.new("char[]", destination_dir.encode("ascii"))

        res = lib.fetchHelmChart(c_repo_url, c_chart_name, c_version, c_destination_dir)
        response = ffi.string(res).decode("utf-8")
        if response != "":
            logger.warning(
                "Dependency helm chart {} and version {}: {}".format(chart_name, version, response)
            )
        else:
            logger.info(
                "Dependency helm chart {} and version {}: successfully fetched".format(chart_name, version)
            )
            logger.info(
                "Dependency helm chart {} and version {}: saved to {}".format(
                    chart_name, version, destination_dir
                )
            )


def initialise_helm_fetch_binding():
    """returns the dl_opened library (.so file) if exists, otherwise None"""
    if platform.system() not in ("Linux", "Darwin"):  # TODO: later add binding for Mac
        return None
    # binding_path is kapitan/dependency_manager/helm/helm_fetch.so
    binding_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "helm/helm_fetch.so")
    if not os.path.exists(binding_path):
        logger.debug("The helm_fetch binding does not exist at {}".format(binding_path))
        return None
    try:
        lib = ffi.dlopen(binding_path)
    except (NameError, OSError) as e:
        raise HelmBindingUnavailableError(
            "There was an error opening helm_fetch.so binding. " "Refer to the exception below:\n" + str(e)
        )
    return lib



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"binary tests"

import base64
import contextlib
from distutils import dir_util
import io
import os
import tempfile
import shutil
import subprocess
import sys
import unittest

from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.utils import directory_hash

REFS_PATH = tempfile.mkdtemp()
BINARY_PATH = "bindist/kapitan-linux-amd64"


@unittest.skipIf(not os.path.exists(BINARY_PATH), "kapitan binary not found")
class BinaryTest(unittest.TestCase):
    def test_cli_secret_validate_targets(self):
        """
        run $ kapitan refs --validate-targets
        expect 0 (success) exit status code
        """
        argv = [
            "kapitan",
            "refs",
            "--validate-targets",
            "--refs-path",
            "examples/kubernetes/refs/targets/",
            "--inventory-path",
            "examples/kubernetes/inventory/",
        ]
        with self.assertRaises(SystemExit) as cm:
            sys.argv = argv
            main()
        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual(cm.exception.code, result.returncode)

    def test_cli_secret_write_reveal_gkms(self):
        """
        run $ kapitan refs --write gkms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [
            BINARY_PATH,
            "refs",
            "--write",
            "gkms:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--key",
            "mock",
        ]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{gkms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content), result.stdout.decode("utf-8"))

        os.remove(test_tag_file)

    def test_cli_secret_write_reveal_awskms(self):
        """
        run $ kapitan refs --write awskms:test_secret
        and $ kapitan refs --reveal
        using mock KMS key
        """
        test_secret_content = "mock"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [
            BINARY_PATH,
            "refs",
            "--write",
            "awskms:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
            "--key",
            "mock",
        ]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{awskms:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content), result.stdout.decode("utf-8"))

        os.remove(test_tag_file)

    def test_cli_secret_write_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [
            BINARY_PATH,
            "refs",
            "--write",
            "base64:test_secret",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        subprocess.run(argv)
        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual("revealing: {}".format(test_secret_content), result.stdout.decode("utf-8"))

        os.remove(test_tag_file)

    def test_cli_secret_write_base64_ref(self):
        """
        run $ kapitan refs --write base64:test_secret --base64
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = "secret_value!"
        test_secret_content_b64 = base64.b64encode(test_secret_content.encode())
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [
            BINARY_PATH,
            "refs",
            "--write",
            "base64:test_secret",
            "--base64",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        subprocess.run(argv)

        test_tag_content = "revealing: ?{base64:test_secret}"
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual(
            "revealing: {}".format(test_secret_content_b64.decode()), result.stdout.decode("utf-8")
        )

        os.remove(test_tag_file)

    def test_cli_secret_subvar_ref(self):
        """
        run $ kapitan refs --write base64:test_secret
        and $ kapitan refs --reveal -f sometest_file
        """
        test_secret_content = """
        var1:
          var2: hello
        var3:
          var4: world
        """
        test_secret_file = tempfile.mktemp()
        with open(test_secret_file, "w") as fp:
            fp.write(test_secret_content)

        argv = [
            BINARY_PATH,
            "refs",
            "--write",
            "base64:test_secret_subvar",
            "-f",
            test_secret_file,
            "--refs-path",
            REFS_PATH,
        ]
        subprocess.run(argv)
        test_tag_content = """
        revealing1: ?{base64:test_secret_subvar@var1.var2}
        revealing2: ?{base64:test_secret_subvar@var3.var4}
        """
        test_tag_file = tempfile.mktemp()
        with open(test_tag_file, "w") as fp:
            fp.write(test_tag_content)
        argv = [BINARY_PATH, "refs", "--reveal", "-f", test_tag_file, "--refs-path", REFS_PATH]

        result = subprocess.run(argv, stdout=subprocess.PIPE)
        expected = """
        revealing1: {}
        revealing2: {}
        """
        self.assertEqual(expected.format("hello", "world"), result.stdout.decode("utf-8"))
        os.remove(test_tag_file)

    def test_cli_searchvar(self):
        """
        run $ kapitan searchvar mysql.replicas
        """
        argv = [
            "kapitan",
            "searchvar",
            "mysql.replicas",
            "--inventory-path",
            "examples/kubernetes/inventory/",
        ]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            sys.argv = argv
            main()

        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)

        self.assertEqual(result.stdout.decode("utf-8"), stdout.getvalue())

    def test_cli_inventory(self):
        """
        run $ kapitan inventory -t minikube-es -F -p cluster
        """
        argv = [
            "kapitan",
            "inventory",
            "-t",
            "minikube-es",
            "-F",
            "-p",
            "cluster",
            "--inventory-path",
            "examples/kubernetes/inventory/",
        ]

        # set stdout as string
        stdout = io.StringIO()
        with contextlib.redirect_stdout(stdout):
            sys.argv = argv
            main()
        argv[0] = BINARY_PATH
        result = subprocess.run(argv, stdout=subprocess.PIPE)
        self.assertEqual(result.stdout.decode("utf-8"), stdout.getvalue())

    def tearDown(self):
        shutil.rmtree(REFS_PATH, ignore_errors=True)


@unittest.skipIf(not os.path.exists(BINARY_PATH), "kapitan binary not found")
class TestTerraformCompile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.chdir(os.getcwd() + "/examples/terraform/")

    def test_cli_terraform_compile(self):
        """
        run $kapitan compile
        """
        sys.argv = ["kapitan", "compile"]
        main()
        compiled_dir_hash = directory_hash(os.getcwd() + "/compiled")
        test_compiled_dir_hash = directory_hash(os.getcwd() + "/../../tests/test_terraform_compiled")
        self.assertEqual(compiled_dir_hash, test_compiled_dir_hash)
        compile_path = tempfile.mkdtemp()
        argv = ["kapitan", "compile"]
        sys.argv = argv
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        argv[0] = "../../" + BINARY_PATH
        argv.extend(["--output-path", compile_path])
        subprocess.run(argv, stdout=subprocess.DEVNULL)
        main_compiled_dir_hash = directory_hash(os.getcwd() + "/compiled")
        binary_compiled_dir_hash = directory_hash(compile_path + "/compiled")
        self.assertEqual(main_compiled_dir_hash, binary_compiled_dir_hash)

    @classmethod
    def tearDownClass(cls):
        os.chdir(os.getcwd() + "/../../")
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), "kapitan binary not found")
class TestKubernetesCompile(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        os.chdir(os.getcwd() + "/examples/kubernetes/")

    def test_cli_kubenetes_compile(self):
        """
        run $kapitan compile
        """
        compile_path = tempfile.mkdtemp()
        argv = ["kapitan", "compile"]
        sys.argv = argv
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        argv[0] = "../../" + BINARY_PATH
        argv.extend(["--output-path", compile_path])
        subprocess.run(argv, stdout=subprocess.DEVNULL)
        main_compiled_dir_hash = directory_hash(os.getcwd() + "/compiled")
        binary_compiled_dir_hash = directory_hash(compile_path + "/compiled")
        self.assertEqual(main_compiled_dir_hash, binary_compiled_dir_hash)

    @classmethod
    def tearDownClass(cls):
        os.chdir(os.getcwd() + "/../../")
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), "kapitan binary not found")
class TestDependencyManager(unittest.TestCase):
    def test_compile_fetch(self):
        temp = tempfile.mkdtemp()
        dir_util.copy_tree("tests/test_resources", temp)
        argv = [
            os.path.join(os.getcwd(), BINARY_PATH),
            "compile",
            "--fetch",
            "--output-path",
            temp,
            "-t",
            "nginx",
            "nginx-dev",
            "monitoring-dev",
            "-p",
            "4",
        ]
        subprocess.run(argv, cwd=temp, stdout=subprocess.DEVNULL)
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "tests")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "acs-engine-autoscaler")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "kapitan-repository")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "components", "source")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "charts", "prometheus")))

    def tearDown(self):
        reset_cache()


@unittest.skipIf(not os.path.exists(BINARY_PATH), "kapitan binary not found")
class TestHelmInputBinary(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join("tests", "test_resources"))

    def test_compile_helm_input(self):
        """compile targets with helm inputs in parallel"""
        temp_main = tempfile.mkdtemp()
        sys.argv = [
            "kapitan",
            "compile",
            "--output-path",
            temp_main,
            "-t",
            "nginx-ingress",
            "nginx-ingress-helm-params",
            "acs-engine-autoscaler",
        ]
        with contextlib.redirect_stdout(io.StringIO()):
            main()
        main_hash = directory_hash(temp_main + "/compiled")

        temp_bin = tempfile.mkdtemp()
        subprocess.run(
            [
                "../../" + BINARY_PATH,
                "compile",
                "--output-path",
                temp_bin,
                "-t",
                "nginx-ingress",
                "nginx-ingress-helm-params",
                "acs-engine-autoscaler",
            ],
            stdout=subprocess.DEVNULL,
        )
        bin_hash = directory_hash(temp_bin + "/compiled")
        self.assertEqual(bin_hash, main_hash)

    @classmethod
    def tearDownClass(cls):
        reset_cache()
        os.chdir("../../")



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import six
import re

class DictPath(object):
    '''
    Represents a path into a nested dictionary.

    Given a dictionary like

      d['foo']['bar'] = 42

    it can be desirable to obtain a reference to the value stored in the
    sub-levels, allowing that value to be accessed and changed. Unfortunately,
    Python provides no easy way to do this, since

      ref = d['foo']['bar']

    does become a reference to the integer 42, but that reference is
    overwritten when one assigns to it. Hence, DictPath represents the path
    into a nested dictionary, and can be "applied to" a dictionary to obtain
    and set values, using a list of keys, or a string representation using
    a delimiter (which can be escaped):

      p = DictPath(':', 'foo:bar')
      p.get_value(d)
      p.set_value(d, 43)

    This is a bit backwards, but the right way around would require support by
    the dict() type.

    The primary purpose of this class within reclass is to cater for parameter
    interpolation, so that a reference such as ${foo:bar} in a parameter value
    may be resolved in the context of the Parameter collections (a nested
    dict).

    If the value is a list, then the "key" is assumed to be and interpreted as
    an integer index:

      d = {'list': [{'one':1},{'two':2}]}
      p = DictPath(':', 'list:1:two')
      p.get_value(d)  → 2

    This heuristic is okay within reclass, because dictionary keys (parameter
    names) will always be strings. Therefore it is okay to interpret each
    component of the path as a string, unless one finds a list at the current
    level down the nested dictionary.
    '''

    def __init__(self, delim, contents=None):
        self._delim = delim

        if contents is None:
            self._parts = []
        elif isinstance(contents, list):
            self._parts = contents
        elif isinstance(contents, six.string_types):
            self._parts = self._split_string(contents)
        elif isinstance(contents, tuple):
            self._parts = list(contents)
        else:
            raise TypeError('DictPath() takes string or list, '\
                                'not %s' % type(contents))

    def __repr__(self):
        return "DictPath(%r, %r)" % (self._delim, str(self))

    def __str__(self):
        return self._delim.join(str(i) for i in self._parts)

    def __eq__(self, other):
        if not (isinstance(other, six.string_types) or
                isinstance(other, self.__class__)):
            return False
        if isinstance(other, six.string_types):
            other = DictPath(self._delim, other)
        return self._parts == other._parts and self._delim == other._delim

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))

    @property
    def path(self):
        return self._parts

    def _get_key(self):
        if len(self._parts) == 0:
            return None
        return self._parts[-1]

    def _get_innermost_container(self, base):
        container = base
        for i in self.path[:-1]:
            if isinstance(container, (list, tuple)):
                container = container[int(i)]
            else:
                container = container[i]
        return container

    def _split_string(self, string):
        return re.split(r'(?<!\\)' + re.escape(self._delim), string)

    def key_parts(self):
        return self._parts[:-1]

    def new_subpath(self, key):
        return DictPath(self._delim, self._parts + [key])

    def get_value(self, base):
        return self._get_innermost_container(base)[self._get_key()]

    def set_value(self, base, value):
        self._get_innermost_container(base)[self._get_key()] = value

    def drop_first(self):
        del self._parts[0]
        return self

    def is_empty(self):
        return len(self._parts) == 0

    def delete(self, base):
        del self._get_innermost_container(base)[self._get_key()]

    def add_subpath(self, key):
        self._parts.append(key)

    def add_ancestor(self, key):
        self._parts.insert(0, key)

    def is_ancestor_of(self, other):
        if len(other._parts) <= len(self._parts):
            return False
        for i in range(len(self._parts)):
            if other._parts[i] != self._parts[i]:
                return False
        return True

    def exists_in(self, container):
        item = container
        for part in self._parts:
            if isinstance(item, (dict, list)):
                if part in item:
                    if isinstance(item, dict):
                        item = item[part]
                    elif isinstance(container, list):
                        item = item[int(part)]
                else:
                    return False
            else:
                if item == self._parts[-1]:
                    return True
                return False
        return True



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"helm input tests"
import os
import sys
import tempfile
import unittest

import yaml
from kapitan.cached import reset_cache
from kapitan.cli import main
from kapitan.inputs.helm import Helm

helm_binding_exists = True
try:
    from kapitan.inputs.helm.helm_binding import (
        ffi,
    )  # this statement will raise ImportError if binding not available
except ImportError:
    helm_binding_exists = False


@unittest.skipUnless(helm_binding_exists, "helm binding is not available")
class HelmInputTest(unittest.TestCase):
    def setUp(self):
        os.chdir(os.path.join("tests", "test_resources"))

    def test_render_chart(self):
        temp_dir = tempfile.mkdtemp()
        chart_path = "charts/acs-engine-autoscaler"
        helm = Helm(None, None, None)
        error_message = helm.render_chart(chart_path, temp_dir)
        self.assertFalse(error_message)
        self.assertTrue(
            os.path.isfile(os.path.join(temp_dir, "acs-engine-autoscaler", "templates", "secrets.yaml"))
        )
        self.assertTrue(
            os.path.isfile(os.path.join(temp_dir, "acs-engine-autoscaler", "templates", "deployment.yaml"))
        )

    def test_error_invalid_char_dir(self):
        chart_path = "non-existent"
        temp_dir = tempfile.mkdtemp()
        helm = Helm(None, None, None)
        error_message = helm.render_chart(chart_path, temp_dir)
        self.assertTrue("no such file or directory" in error_message)

    def test_compile_chart(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "acs-engine-autoscaler"]
        main()
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    temp,
                    "compiled",
                    "acs-engine-autoscaler",
                    "acs-engine-autoscaler",
                    "templates",
                    "secrets.yaml",
                )
            )
        )

    def test_compile_subcharts(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "istio"]
        main()
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "istio", "istio", "charts")))
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "istio", "istio", "templates")))

    def test_compile_multiple_targets(self):
        temp = tempfile.mkdtemp()
        sys.argv = [
            "kapitan",
            "compile",
            "--output-path",
            temp,
            "-t",
            "acs-engine-autoscaler",
            "nginx-ingress",
            "-p",
            "2",
        ]
        main()
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    temp,
                    "compiled",
                    "acs-engine-autoscaler",
                    "acs-engine-autoscaler",
                    "templates",
                    "secrets.yaml",
                )
            )
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(
                    temp, "compiled", "nginx-ingress", "nginx-ingress", "templates", "clusterrolebinding.yaml"
                )
            )
        )

    def test_compile_multiple_charts_per_target(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-istio"]
        main()
        self.assertTrue(os.path.isdir(os.path.join(temp, "compiled", "nginx-istio", "istio", "templates")))
        self.assertTrue(
            os.path.isdir(os.path.join(temp, "compiled", "nginx-istio", "nginx-ingress", "templates"))
        )

    def test_compile_with_helm_values(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-ingress"]
        main()
        controller_deployment_file = os.path.join(
            temp, "compiled", "nginx-ingress", "nginx-ingress", "templates", "controller-deployment.yaml"
        )
        self.assertTrue(os.path.isfile(controller_deployment_file))
        with open(controller_deployment_file, "r") as fp:
            manifest = yaml.safe_load(fp.read())
            name = manifest["metadata"]["name"]
            self.assertEqual(name, "-nginx-ingress-my-controller")

    def test_compile_with_helm_values_files(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "monitoring-dev", "monitoring-prd"]
        main()
        dev_server_deployment_file = os.path.join(
            temp, "compiled", "monitoring-dev", "prometheus", "templates", "server-deployment.yaml"
        )
        prd_server_deployment_file = os.path.join(
            temp, "compiled", "monitoring-prd", "prometheus", "templates", "server-deployment.yaml"
        )

        self.assertTrue(os.path.isfile(dev_server_deployment_file))
        with open(dev_server_deployment_file, "r") as fp:
            manifest = yaml.safe_load(fp.read())
            name = manifest["metadata"]["name"]
            self.assertEqual(name, "prometheus-dev-server")

        self.assertTrue(os.path.isfile(prd_server_deployment_file))
        with open(prd_server_deployment_file, "r") as fp:
            manifest = yaml.safe_load(fp.read())
            name = manifest["metadata"]["name"]
            self.assertEqual(name, "prometheus-prd-server")

    def test_compile_with_helm_params(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-ingress-helm-params"]
        with open("inventory/targets/nginx-ingress-helm-params.yml", "r") as fp:
            manifest = yaml.safe_load(fp.read())
            helm_params = manifest["parameters"]["kapitan"]["compile"][0]["helm_params"]
            release_name = helm_params["release_name"]
            namespace = helm_params["namespace"]

        main()
        controller_deployment_file = os.path.join(
            temp,
            "compiled",
            "nginx-ingress-helm-params",
            "nginx-ingress",
            "templates",
            "controller-deployment.yaml",
        )

        self.assertTrue(os.path.isfile(controller_deployment_file))
        with open(controller_deployment_file, "r") as fp:
            manifest = yaml.safe_load(fp.read())
            container = manifest["spec"]["template"]["spec"]["containers"][0]
            property = container["args"][4]
            self.assertEqual(
                property, "--configmap={}/{}".format(namespace, release_name + "-nginx-ingress-my-controller")
            )

    def test_compile_with_refs(self):
        temp = tempfile.mkdtemp()
        sys.argv = ["kapitan", "compile", "--output-path", temp, "-t", "nginx-ingress", "--reveal"]
        main()
        controller_deployment_file = os.path.join(
            temp, "compiled", "nginx-ingress", "nginx-ingress", "templates", "controller-deployment.yaml"
        )
        self.assertTrue(os.path.isfile(controller_deployment_file))
        with open(controller_deployment_file, "r") as fp:
            manifest = yaml.safe_load(fp.read())
            args = next(
                iter(
                    c["args"]
                    for c in manifest["spec"]["template"]["spec"]["containers"]
                    if c["name"] == "nginx-ingress-my-controller"
                )
            )
            self.assertIn("--election-id=super_secret_ID", args)

    def tearDown(self):
        os.chdir("../../")
        reset_cache()



from reclass.settings import Settings
from reclass.values.value import Value
from reclass.values.compitem import CompItem
from reclass.values.scaitem import ScaItem
from reclass.values.valuelist import ValueList
from reclass.values.listitem import ListItem
from reclass.values.dictitem import DictItem
from reclass.values.item import ContainerItem
from reclass.values.item import ItemWithReferences
import unittest
from mock import MagicMock

SETTINGS = Settings()


class TestItemWithReferences(unittest.TestCase):

    def test_assembleRef_allrefs(self):
        phonyitem = MagicMock()
        phonyitem.has_references = True
        phonyitem.get_references = lambda *x: [1]

        iwr = ItemWithReferences([phonyitem], {})

        self.assertEquals(iwr.get_references(), [1])
        self.assertTrue(iwr.allRefs)

    def test_assembleRef_partial(self):
        phonyitem = MagicMock()
        phonyitem.has_references = True
        phonyitem.allRefs = False
        phonyitem.get_references = lambda *x: [1]

        iwr = ItemWithReferences([phonyitem], {})

        self.assertEquals(iwr.get_references(), [1])
        self.assertFalse(iwr.allRefs)


class TestContainerItem(unittest.TestCase):

    def test_render(self):
        container = ContainerItem('foo', SETTINGS)

        self.assertEquals(container.render(None, None), 'foo')

if __name__ == '__main__':
    unittest.main()



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"awskms secrets module"

import base64
import boto3

from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend
from kapitan import cached
from kapitan.errors import KapitanError


class AWSKMSError(KapitanError):
    """Generic AWS KMS errors"""

    pass


def awskms_obj():
    if not cached.awskms_obj:
        cached.awskms_obj = boto3.session.Session().client("kms")
    return cached.awskms_obj


class AWSKMSSecret(Base64Ref):
    def __init__(self, data, key, encrypt=True, encode_base64=False, **kwargs):
        """
        encrypts data with key
        set encode_base64 to True to base64 encode data before encrypting and writing
        set encrypt to False if loading data that is already encrypted and base64
        """
        if encrypt:
            self._encrypt(data, key, encode_base64)
            if encode_base64:
                kwargs["encoding"] = "base64"
        else:
            self.data = data
            self.key = key
        super().__init__(self.data, **kwargs)
        self.type_name = "awskms"

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new AWSKMSSecret from data and ref_params: target_name
        key will be grabbed from the inventory via target_name
        """
        try:
            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv["nodes"].get(target_name, None)
            if target_inv is None:
                raise ValueError("target_inv not set")

            key = target_inv["parameters"]["kapitan"]["secrets"]["awskms"]["key"]
            return cls(data, key, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create AWSKMSSecret: target_name missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        returns decrypted data
        raises AWSKMSError if decryption fails
        """
        # can't use super().reveal() as we want bytes
        ref_data = base64.b64decode(self.data)
        return self._decrypt(ref_data)

    def update_key(self, key):
        """
        re-encrypts data with new key, respects original encoding
        returns True if key is different and secret is updated, False otherwise
        """
        if key == self.key:
            return False

        data_dec = self.reveal()
        encode_base64 = self.encoding == "base64"
        if encode_base64:
            data_dec = base64.b64decode(data_dec).decode()
        self._encrypt(data_dec, key, encode_base64)
        self.data = base64.b64encode(self.data).decode()
        return True

    def _encrypt(self, data, key, encode_base64):
        """
        encrypts data
        set encode_base64 to True to base64 encode data before writing
        """
        assert isinstance(key, str)
        _data = data
        self.encoding = "original"
        if encode_base64:
            _data = base64.b64encode(data.encode())
            self.encoding = "base64"
        else:
            # To guarantee _data is bytes
            if isinstance(data, str):
                _data = data.encode()
        try:
            ciphertext = ""
            # Mocking encrypted response for tests
            if key == "mock":
                ciphertext = base64.b64encode("mock".encode())
            else:
                response = awskms_obj().encrypt(KeyId=key, Plaintext=_data)
                ciphertext = base64.b64encode(response["CiphertextBlob"])
            self.data = ciphertext
            self.key = key

        except Exception as e:
            raise AWSKMSError(e)

    def _decrypt(self, data):
        """decrypt data"""
        try:
            plaintext = ""
            # Mocking decrypted response for tests
            if self.key == "mock":
                plaintext = "mock".encode()
            else:
                response = awskms_obj().decrypt(CiphertextBlob=base64.b64decode(data))
                plaintext = response["Plaintext"]

            return plaintext.decode()

        except Exception as e:
            raise AWSKMSError(e)

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {"data": self.data, "encoding": self.encoding, "key": self.key, "type": self.type_name}


class AWSKMSBackend(Base64RefBackend):
    def __init__(self, path, ref_type=AWSKMSSecret, **ref_kwargs):
        "init AWSKMSBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = "awskms"



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"gpg secrets module"

import base64
import logging
import time

import gnupg
from kapitan import cached
from kapitan.errors import KapitanError
from kapitan.refs.base import RefError
from kapitan.refs.base64 import Base64Ref, Base64RefBackend

logger = logging.getLogger(__name__)


# XXX only use this for testing!
# pass custom kwargs to gpg encrypt()/decrypt()
GPG_KWARGS = {}

# XXX only use this for testing!
# pass custom fingerprints within from_params()
GPG_TARGET_FINGERPRINTS = {}


class GPGError(Exception):
    """Generic GPG errors"""

    pass


def gpg_obj(*args, **kwargs):
    if not cached.gpg_obj:
        cached.gpg_obj = gnupg.GPG(*args, **kwargs)
    return cached.gpg_obj


class GPGSecret(Base64Ref):
    def __init__(self, data, recipients, encrypt=True, encode_base64=False, **kwargs):
        """
        encrypts data for recipients
        set encode_base64 to True to base64 encode data before encrypting and writing
        set encrypt to False if loading data that is already encrypted and base64
        if fingerprint key is not set in recipients, the first non-expired fingerprint will be used
        if fingerprint is set, there will be no name based lookup
        """
        fingerprints = lookup_fingerprints(recipients)
        if encrypt:
            self._encrypt(data, fingerprints, encode_base64)
            if encode_base64:
                kwargs["encoding"] = "base64"
        else:
            self.data = data
            self.recipients = [{"fingerprint": f} for f in fingerprints]  # TODO move to .load() method
        super().__init__(self.data, **kwargs)
        self.type_name = "gpg"

    @classmethod
    def from_params(cls, data, ref_params):
        """
        Return new GPPSecret from data and ref_params: target_name
        recipients will be grabbed from the inventory via target_name
        """
        try:
            # XXX only used for testing
            if GPG_TARGET_FINGERPRINTS:
                _fingerprints = [{"fingerprint": v} for _, v in GPG_TARGET_FINGERPRINTS.items()]
                return cls(data, _fingerprints, **ref_params.kwargs)

            target_name = ref_params.kwargs["target_name"]
            if target_name is None:
                raise ValueError("target_name not set")

            target_inv = cached.inv["nodes"].get(target_name, None)
            if target_inv is None:
                raise ValueError("target_inv not set")

            if "secrets" not in target_inv["parameters"]["kapitan"]:
                raise KapitanError(
                    f"parameters.kapitan.secrets not defined in inventory of target {target_name}"
                )

            recipients = target_inv["parameters"]["kapitan"]["secrets"]["gpg"]["recipients"]

            return cls(data, recipients, **ref_params.kwargs)
        except KeyError:
            raise RefError("Could not create GPGSecret: target_name missing")

    @classmethod
    def from_path(cls, ref_full_path, **kwargs):
        return super().from_path(ref_full_path, encrypt=False, **kwargs)

    def reveal(self):
        """
        returns decrypted data
        raises GPGError if decryption fails
        """
        # can't use super().reveal() as we want bytes
        ref_data = base64.b64decode(self.data)
        return self._decrypt(ref_data)

    def update_recipients(self, recipients):
        """
        re-encrypts data with new recipients, respects original encoding
        returns True if recipients are different and secret is updated, False otherwise
        """
        fingerprints = lookup_fingerprints(recipients)
        if set(fingerprints) == set([r["fingerprint"] for r in self.recipients]):
            return False

        data_dec = self.reveal()
        encode_base64 = self.encoding == "base64"
        if encode_base64:
            data_dec = base64.b64decode(data_dec).decode()
        self._encrypt(data_dec, fingerprints, encode_base64)
        self.data = base64.b64encode(self.data).decode()
        return True

    def _encrypt(self, data, fingerprints, encode_base64):
        """
        encrypts data
        set encode_base64 to True to base64 encode data before writing
        """
        assert isinstance(fingerprints, list)
        _data = data
        self.encoding = "original"
        if encode_base64:
            _data = base64.b64encode(data.encode())
            self.encoding = "base64"
        enc = gpg_obj().encrypt(_data, fingerprints, sign=True, armor=False, **GPG_KWARGS)
        if enc.ok:
            self.data = enc.data
            self.recipients = [{"fingerprint": f} for f in fingerprints]
        else:
            raise GPGError(enc.status)

    def _decrypt(self, data):
        """decrypt data"""
        dec = gpg_obj().decrypt(data, **GPG_KWARGS)
        if dec.ok:
            return dec.data.decode()
        else:
            raise GPGError(dec.status)

    def dump(self):
        """
        Returns dict with keys/values to be serialised.
        """
        return {
            "data": self.data,
            "encoding": self.encoding,
            "recipients": self.recipients,
            "type": self.type_name,
        }


class GPGBackend(Base64RefBackend):
    def __init__(self, path, ref_type=GPGSecret, **ref_kwargs):
        "init GPGBackend ref backend type"
        super().__init__(path, ref_type, **ref_kwargs)
        self.type_name = "gpg"


def lookup_fingerprints(recipients):
    """returns a list of fingerprints for recipients obj"""
    lookedup = []
    for recipient in recipients:
        fingerprint = recipient.get("fingerprint")
        name = recipient.get("name")
        if fingerprint is None:
            lookedup_fingerprint = fingerprint_non_expired(name)
            lookedup.append(lookedup_fingerprint)
        else:
            # If fingerprint already set, don't lookup and reuse
            lookedup.append(fingerprint)

    # Remove duplicates, sort and return fingerprints list
    return sorted(set(lookedup))


def fingerprint_non_expired(recipient_name):
    """returns first non-expired key fingerprint for recipient_name"""
    try:
        keys = gpg_obj().list_keys(keys=(recipient_name,))
        for key in keys:
            if "expires" not in key:
                logger.info(
                    "Invalid dictionary structure for key for recipient: %s with fingerprint: %s",
                    recipient_name,
                    key["fingerprint"],
                )
                continue

            # if 'expires' is indefinite (meaning it is an empty string) OR
            # if 'expires' key is set and time is in the future, return
            if (not key["expires"]) or (time.time() < int(key["expires"])):
                return key["fingerprint"]
            else:
                logger.debug(
                    f"Key for recipient: {recipient_name} with fingerprint: {key['fingerprint']} has expired, skipping"
                )
        raise GPGError(f"Could not find valid key for recipient: {recipient_name}")
    except IndexError as iexp:
        raise iexp



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"cached module"

inv = {}
inv_cache = {}
gpg_obj = None
gkms_obj = None
awskms_obj = None
dot_kapitan = {}
ref_controller_obj = None
revealer_obj = None
args = {}  # args won't need resetting


def reset_cache():
    global inv, inv_cache, gpg_obj, gkms_obj, awskms_obj, dot_kapitan, ref_controller_obj, revealer_obj
    inv = {}
    inv_cache = {}
    gpg_obj = None
    gkms_obj = None
    awskms_obj = None
    dot_kapitan = {}
    ref_controller_obj = None
    revealer_obj = None



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import copy

from six import iteritems, next

from .parameters import Parameters
from reclass.errors import ResolveError
from reclass.values.value import Value
from reclass.values.valuelist import ValueList
from reclass.utils.dictpath import DictPath

class Exports(Parameters):

    def __init__(self, mapping, settings, uri):
        super(Exports, self).__init__(mapping, settings, uri)

    def delete_key(self, key):
        self._base.pop(key, None)
        self._unrendered.pop(key, None)

    def overwrite(self, other):
        overdict = {'~' + key: value for (key, value) in iteritems(other)}
        self.merge(overdict)

    def interpolate_from_external(self, external):
        while len(self._unrendered) > 0:
            path, v = next(iteritems(self._unrendered))
            value = path.get_value(self._base)
            if isinstance(value, (Value, ValueList)):
                external._interpolate_references(path, value, None)
                new = self._interpolate_render_from_external(external._base, path, value)
                path.set_value(self._base, new)
                del self._unrendered[path]
            else:
                # references to lists and dicts are only deepcopied when merged
                # together so it's possible a value with references in a referenced
                # list or dict has already been rendered
                del self._unrendered[path]

    def interpolate_single_from_external(self, external, query):
        for r in query.get_inv_references():
            self._interpolate_single_path_from_external(r, external, query)

    def _interpolate_single_path_from_external(self, mainpath, external, query):
        required = self._get_required_paths(mainpath)
        while len(required) > 0:
            while len(required) > 0:
                path, v = next(iteritems(required))
                value = path.get_value(self._base)
                if isinstance(value, (Value, ValueList)):
                    try:
                        external._interpolate_references(path, value, None)
                        new = self._interpolate_render_from_external(external._base, path, value)
                        path.set_value(self._base, new)
                    except ResolveError as e:
                        if query.ignore_failed_render():
                            path.delete(self._base)
                        else:
                            raise
                del required[path]
                del self._unrendered[path]
            required = self._get_required_paths(mainpath)

    def _get_required_paths(self, mainpath):
        paths = {}
        path = DictPath(self._settings.delimiter)
        for i in mainpath.key_parts():
            path.add_subpath(i)
            if path in self._unrendered:
                paths[path] = True
        for i in self._unrendered:
            if mainpath.is_ancestor_of(i) or mainpath == i:
                paths[i] = True
        return paths

    def _interpolate_render_from_external(self, context, path, value):
        try:
            new = value.render(context, None)
        except ResolveError as e:
            e.context = path
            raise
        if isinstance(new, dict):
            new = self._render_simple_dict(new, path)
        elif isinstance(new, list):
            new = self._render_simple_list(new, path)
        return new



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"kapitan resources"

import base64
import errno
import gzip
import io
import json
import logging
import os
import sys
from functools import partial

import jsonschema
import kapitan.cached as cached
import yaml
from kapitan import __file__ as kapitan_install_path
from kapitan import cached as cached
from kapitan.errors import CompileError, InventoryError, KapitanError
from kapitan.utils import PrettyDumper, deep_get, flatten_dict, render_jinja2_file, sha256_string

import reclass
import reclass.core
from reclass.errors import NotFoundError, ReclassException

logger = logging.getLogger(__name__)

try:
    from yaml import CSafeLoader as YamlLoader
except ImportError:
    from yaml import SafeLoader as YamlLoader


def resource_callbacks(search_paths):
    """
    Returns a dict with all the functions to be used
    on the native_callbacks keyword on jsonnet
    search_paths can be used by the native functions to access files
    """

    return {
        "jinja2_render_file": (("name", "ctx"), partial(jinja2_render_file, search_paths)),
        "inventory": (("target", "inv_path"), partial(inventory, search_paths)),
        "file_read": (("name",), partial(read_file, search_paths)),
        "file_exists": (("name",), partial(file_exists, search_paths)),
        "dir_files_list": (("name",), partial(dir_files_list, search_paths)),
        "dir_files_read": (("name",), partial(dir_files_read, search_paths)),
        "sha256_string": (("obj",), sha256_string),
        "gzip_b64": (("obj",), gzip_b64),
        "yaml_dump": (("obj",), yaml_dump),
        "yaml_dump_stream": (("obj",), yaml_dump_stream),
        "yaml_load": (("name",), partial(yaml_load, search_paths)),
        "yaml_load_stream": (("name",), partial(yaml_load_stream, search_paths)),
        "jsonschema_validate": (("obj", "schema_obj"), jsonschema_validate),
    }


def jsonschema_validate(obj, schema_obj):
    """
    validates obj with a schema_obj jsonschema
    returns an object with keys:
    - valid (true/false)
    - reason (will have jsonschema validation error when not valid)
    """
    _obj = json.loads(obj)
    _schema_obj = json.loads(schema_obj)
    try:
        jsonschema.validate(_obj, _schema_obj, format_checker=jsonschema.FormatChecker())
        return {"valid": True, "reason": ""}
    except jsonschema.ValidationError as e:
        return {"valid": False, "reason": "" + str(e)}


def yaml_dump(obj):
    """Dumps jsonnet obj as yaml"""
    _obj = json.loads(obj)
    return yaml.safe_dump(_obj, default_flow_style=False)


def yaml_dump_stream(obj):
    """Dumps jsonnet obj as yaml stream"""
    _obj = json.loads(obj)
    return yaml.safe_dump_all(_obj, default_flow_style=False)


def gzip_b64(obj):
    """returns base64-encoded gzip-compressed obj"""
    obj_bytes = obj.encode("UTF-8")
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode="wb", compresslevel=9, mtime=0) as f:
        f.write(obj_bytes)
    compressed_obj = buf.getvalue()
    return base64.b64encode(compressed_obj).decode("UTF-8")


def jinja2_render_file(search_paths, name, ctx):
    """
    Render jinja2 file name with context ctx.
    search_paths is used to find the file name
    as there is a limitation with jsonnet's native_callback approach:
    one can't access the current directory being evaluated
    """
    ctx = json.loads(ctx)
    _full_path = ""

    for path in search_paths:
        _full_path = os.path.join(path, name)
        logger.debug("jinja2_render_file trying file %s", _full_path)
        if os.path.exists(_full_path):
            logger.debug("jinja2_render_file found file at %s", _full_path)
            try:
                return render_jinja2_file(_full_path, ctx, search_paths=search_paths)
            except Exception as e:
                raise CompileError("Jsonnet jinja2 failed to render {}: {}".format(_full_path, e))

    raise IOError("jinja2 failed to render, could not find file: {}".format(_full_path))


def yaml_load(search_paths, name):
    """returns content of yaml file as json string"""
    for path in search_paths:
        _full_path = os.path.join(path, name)
        logger.debug("yaml_load trying file %s", _full_path)
        if os.path.exists(_full_path) and (name.endswith(".yml") or name.endswith(".yaml")):
            logger.debug("yaml_load found file at %s", _full_path)
            try:
                with open(_full_path) as f:
                    return json.dumps(yaml.safe_load(f.read()))
            except Exception as e:
                raise CompileError("Parse yaml failed to parse {}: {}".format(_full_path, e))

    raise IOError("could not find any input yaml file: {}".format(_full_path))


def yaml_load_stream(search_paths, name):
    """returns contents of yaml file as generator"""
    for path in search_paths:
        _full_path = os.path.join(path, name)
        logger.debug("yaml_load_stream trying file %s", _full_path)
        if os.path.exists(_full_path) and (name.endswith(".yml") or name.endswith(".yaml")):
            logger.debug("yaml_load_stream found file at %s", _full_path)
            try:
                with open(_full_path) as f:
                    _obj = yaml.load_all(f.read(), Loader=yaml.SafeLoader)
                    return json.dumps(list(_obj))
            except Exception as e:
                raise CompileError("Parse yaml failed to parse {}: {}".format(_full_path, e))

    raise IOError("could not find any input yaml file: {}".format(_full_path))


def read_file(search_paths, name):
    """return content of file in name"""
    for path in search_paths:
        full_path = os.path.join(path, name)
        logger.debug("read_file trying file %s", full_path)
        if os.path.exists(full_path):
            logger.debug("read_file found file at %s", full_path)
            with io.open(full_path, newline="") as f:
                return f.read()

    raise IOError("Could not find file {}".format(name))


def file_exists(search_paths, name):
    """returns an object with keys:
    - exists (true/false)
    - path (will have the full path where the file was found)"""
    for path in search_paths:
        full_path = os.path.join(path, name)
        logger.debug("file_exists trying file %s", full_path)
        if os.path.exists(full_path):
            logger.debug("file_exists found file at %s", full_path)
            return {"exists": True, "path": full_path}

    return {"exists": False, "path": ""}


def dir_files_list(search_paths, name):
    """returns list of files in a dir"""
    for path in search_paths:
        full_path = os.path.join(path, name)
        logger.debug("dir_files_list trying directory %s", full_path)
        if os.path.exists(full_path):
            return [f for f in os.listdir(full_path) if os.path.isfile(os.path.join(full_path, f))]
    raise IOError("Could not find folder {}".format(name))


def dir_files_read(search_paths, name):
    """returns an object with key:
    - file_name (contents of the file)"""
    for path in search_paths:
        full_path = os.path.join(path, name)
        logger.debug("dir_files_list trying directory %s", full_path)
        if os.path.exists(full_path):
            return {f: read_file([full_path], f) for f in dir_files_list([full_path], "")}


def search_imports(cwd, import_str, search_paths):
    """
    This is only to be used as a callback for the jsonnet API!
    - cwd is the import context $CWD,
    - import_str is the "bla.jsonnet" in 'import "bla.jsonnet"'
    - search_paths is the location where to look for import_str if not in cwd
    The only supported parameters are cwd and import_str, so search_paths
    needs to be closured.
    """
    basename = os.path.basename(import_str)
    full_import_path = os.path.join(cwd, import_str)

    if not os.path.exists(full_import_path):
        # if import_str not found, search in install_path
        install_path = os.path.dirname(kapitan_install_path)
        _full_import_path = os.path.join(install_path, import_str)
        # if found, set as full_import_path
        if os.path.exists(_full_import_path):
            full_import_path = _full_import_path
            logger.debug(f"import_str: {import_str} found in search_path: {install_path}")
        else:
            # if import_str not found, search in search_paths
            for path in search_paths:
                _full_import_path = os.path.join(path, import_str)
                # if found, set as full_import_path
                if os.path.exists(_full_import_path):
                    full_import_path = _full_import_path
                    logger.debug(f"import_str: {import_str} found in search_path: {path}")
                    break

    # if the above search did not find anything, let jsonnet error
    # with a non existent import
    normalised_path = os.path.normpath(full_import_path)

    logger.debug("cwd:%s import_str:%s basename:%s -> norm:%s", cwd, import_str, basename, normalised_path)

    normalised_path_content = ""
    with open(normalised_path) as f:
        normalised_path_content = f.read()

    return normalised_path, normalised_path_content


def inventory(search_paths, target, inventory_path="inventory/"):
    """
    Reads inventory (set by inventory_path) in search_paths.
    set nodes_uri to change reclass nodes_uri the default value
    set target to None to return all target in the inventory
    Returns a dictionary with the inventory for target
    """

    full_inv_path = ""
    inv_path_exists = False
    for path in search_paths:
        full_inv_path = os.path.join(path, inventory_path)
        if os.path.exists(full_inv_path):
            inv_path_exists = True
            break

    if not inv_path_exists:
        raise InventoryError(f"Inventory not found in search paths: {search_paths}")

    if target is None:
        return inventory_reclass(full_inv_path)["nodes"]

    return inventory_reclass(full_inv_path)["nodes"][target]


def generate_inventory(args):
    if args.pattern and args.target_name == "":
        parser.error("--pattern requires --target_name")
    try:
        inv = inventory_reclass(args.inventory_path)
        if args.target_name != "":
            inv = inv["nodes"][args.target_name]
            if args.pattern != "":
                pattern = args.pattern.split(".")
                inv = deep_get(inv, pattern)
        if args.flat:
            inv = flatten_dict(inv)
            yaml.dump(inv, sys.stdout, width=10000, default_flow_style=False)
        else:
            yaml.dump(inv, sys.stdout, Dumper=PrettyDumper, default_flow_style=False)
    except Exception as e:
        if not isinstance(e, KapitanError):
            logger.exception("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
        sys.exit(1)


def inventory_reclass(inventory_path):
    """
    Runs a reclass inventory in inventory_path
    (same output as running ./reclass.py -b inv_base_uri/ --inventory)
    Will attempt to read reclass config from 'reclass-config.yml' otherwise
    it will failback to the default config.
    Returns a reclass style dictionary
    """

    if not cached.inv:
        reclass_config = {
            "storage_type": "yaml_fs",
            "inventory_base_uri": inventory_path,
            "nodes_uri": os.path.join(inventory_path, "targets"),
            "classes_uri": os.path.join(inventory_path, "classes"),
            "compose_node_name": False,
            "allow_none_override": True,
        }

        try:
            cfg_file = os.path.join(inventory_path, "reclass-config.yml")
            with open(cfg_file) as reclass_cfg:
                reclass_config = yaml.load(reclass_cfg, Loader=YamlLoader)
                # normalise relative nodes_uri and classes_uri paths
                for uri in ("nodes_uri", "classes_uri"):
                    uri_val = reclass_config.get(uri)
                    uri_path = os.path.join(inventory_path, uri_val)
                    normalised_path = os.path.normpath(uri_path)
                    reclass_config.update({uri: normalised_path})
                logger.debug(f"Using reclass inventory config at: {cfg_file}")
        except IOError as ex:
            # If file does not exist, ignore
            if ex.errno == errno.ENOENT:
                logger.debug("Using reclass inventory config defaults")

        try:
            storage = reclass.get_storage(
                reclass_config["storage_type"],
                reclass_config["nodes_uri"],
                reclass_config["classes_uri"],
                reclass_config["compose_node_name"],
            )
            class_mappings = reclass_config.get("class_mappings")  # this defaults to None (disabled)
            _reclass = reclass.core.Core(storage, class_mappings, reclass.settings.Settings(reclass_config))

            cached.inv = _reclass.inventory()
        except ReclassException as e:
            if isinstance(e, NotFoundError):
                logger.error("Inventory reclass error: inventory not found")
            else:
                logger.error("Inventory reclass error: %s", e.message)
            raise InventoryError(e.message)

    return cached.inv



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"init tests"

import logging
import os
import sys
import tempfile
import unittest

from kapitan.cached import reset_cache
from kapitan.cli import main

logging.basicConfig(level=logging.CRITICAL, format="%(message)s")
logger = logging.getLogger(__name__)


class InitTest(unittest.TestCase):
    def test_init(self):
        with tempfile.TemporaryDirectory() as tmp_dir:
            sys.argv = ["kapitan", "init", "--directory", tmp_dir]
            main()

            template_dir = os.path.join(os.getcwd(), "kapitan", "inputs", "templates")

            diff_files = []

            for root, dirs, files in os.walk(tmp_dir):
                diff_files += files

            for root, dirs, files in os.walk(template_dir):
                for f in files:
                    if f in diff_files:
                        diff_files.remove(f)

            self.assertEqual(len(diff_files), 0)

            # check that generated directory compiles
            prevdir = os.getcwd()
            os.chdir(tmp_dir)
            sys.argv = ["kapitan", "compile"]
            main()

            os.chdir(prevdir)
            reset_cache()



# -*- coding: utf-8
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os

class NameMangler:
    @staticmethod
    def nodes(relpath, name):
        # nodes are identified just by their basename, so
        # no mangling required
        return relpath, name

    @staticmethod
    def composed_nodes(relpath, name):
        if relpath == '.' or relpath == '':
            # './' is converted to None
            return None, name
        parts = relpath.split(os.path.sep)
        if parts[0].startswith("_"):
            return relpath, name
        parts.append(name)
        return relpath, '.'.join(parts)

    @staticmethod
    def classes(relpath, name):
        if relpath == '.' or relpath == '':
            # './' is converted to None
            return None, name
        parts = relpath.split(os.path.sep)
        if name != 'init':
            # "init" is the directory index, so only append the basename
            # to the path parts for all other filenames. This has the
            # effect that data in file "foo/init.yml" will be registered
            # as data for class "foo", not "foo.init"
            parts.append(name)
        return relpath, '.'.join(parts)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import sys
from importlib.util import module_from_spec, spec_from_file_location

import inspect
import yaml
from addict import Dict
from kapitan.errors import CompileError
from kapitan.inputs.base import CompiledFile, InputType
from kapitan.resources import inventory as inventory_func
from kapitan.utils import prune_empty

logger = logging.getLogger(__name__)
inventory = None
inventory_global = None
search_paths = None


def module_from_path(path, check_name=None):
    """
    loads python module in path
    set check_name to verify module name against spec name
    returns tuple with module and spec
    """

    if not os.path.isdir(path):
        raise FileNotFoundError("path: {} must be an existing directory".format(path))

    module_name = os.path.basename(os.path.normpath(path))
    init_path = os.path.join(path, "__init__.py")
    spec = spec_from_file_location("kadet_component_{}".format(module_name), init_path)
    mod = module_from_spec(spec)

    if spec is None:
        raise ModuleNotFoundError("Could not load module in path {}".format(path))
    if check_name is not None and check_name != module_name:
        raise ModuleNotFoundError(
            "Module name {} does not match check_name {}".format(module_name, check_name)
        )

    return mod, spec


def load_from_search_paths(module_name):
    """
    loads and executes python module with module_name from search paths
    returns module
    """
    for path in search_paths:
        try:
            _path = os.path.join(path, module_name)
            mod, spec = module_from_path(_path, check_name=module_name)
            spec.loader.exec_module(mod)
            return mod
        except (ModuleNotFoundError, FileNotFoundError):
            pass
    raise ModuleNotFoundError("Could not load module name {}".format(module_name))


class Kadet(InputType):
    def __init__(self, compile_path, search_paths, ref_controller):
        super().__init__("kadet", compile_path, search_paths, ref_controller)
        self.input_params = {}

    def set_input_params(self, input_params):
        self.input_params = input_params

    def compile_file(self, file_path, compile_path, ext_vars, **kwargs):
        """
        Write file_path (kadet evaluated) items as files to compile_path.
        ext_vars is not used in Kadet
        kwargs:
            output: default 'yaml', accepts 'json'
            prune: default False
            reveal: default False, set to reveal refs on compile
            target_name: default None, set to current target being compiled
            indent: default 2
        """
        output = kwargs.get("output", "yaml")
        prune = kwargs.get("prune", False)
        reveal = kwargs.get("reveal", False)
        target_name = kwargs.get("target_name", None)
        indent = kwargs.get("indent", 2)

        input_params = self.input_params
        # set compile_path allowing kadet functions to have context on where files
        # are being compiled on the current kapitan run
        input_params["compile_path"] = compile_path
        # reset between each compile if kadet component is used multiple times
        self.input_params = {}

        # These will be updated per target
        # XXX At the moment we have no other way of setting externals for modules...
        global search_paths
        search_paths = self.search_paths
        global inventory
        inventory = lambda: Dict(inventory_func(self.search_paths, target_name))  # noqa E731
        global inventory_global
        inventory_global = lambda: Dict(inventory_func(self.search_paths, None))  # noqa E731

        kadet_module, spec = module_from_path(file_path)
        sys.modules[spec.name] = kadet_module
        spec.loader.exec_module(kadet_module)
        logger.debug("Kadet.compile_file: spec.name: %s", spec.name)

        kadet_arg_spec = inspect.getfullargspec(kadet_module.main)
        logger.debug("Kadet main args: {}".format(kadet_arg_spec.args))

        if len(kadet_arg_spec.args) == 1:
            output_obj = kadet_module.main(input_params).to_dict()
        elif len(kadet_arg_spec.args) == 0:
            output_obj = kadet_module.main().to_dict()
        else:
            raise ValueError(f"Kadet {spec.name} main parameters not equal to 1 or 0")

        if prune:
            output_obj = prune_empty(output_obj)

        for item_key, item_value in output_obj.items():
            # write each item to disk
            if output == "json":
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_json(item_value)
            elif output in ["yml", "yaml"]:
                file_path = os.path.join(compile_path, "%s.%s" % (item_key, output))
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write_yaml(item_value)
            elif output == "plain":
                file_path = os.path.join(compile_path, "%s" % item_key)
                with CompiledFile(
                    file_path,
                    self.ref_controller,
                    mode="w",
                    reveal=reveal,
                    target_name=target_name,
                    indent=indent,
                ) as fp:
                    fp.write(item_value)
            else:
                raise ValueError(
                    f"Output type defined in inventory for {file_path} is neither 'json', 'yaml' nor 'plain'"
                )
            logger.debug("Pruned output for: %s", file_path)

    def default_output_type(self):
        return "yaml"


class BaseObj(object):
    def __init__(self, **kwargs):
        """
        returns a BaseObj
        kwargs will be save into self.kwargs
        values in self.root are returned as dict via self.to_dict()
        """
        self.root = Dict()
        self.kwargs = Dict(kwargs)
        self.new()
        self.body()

    @classmethod
    def from_json(cls, file_path):
        """
        returns a BaseObj initialised with json content
        from file_path
        """
        with open(file_path) as fp:
            json_obj = json.load(fp)
            return cls.from_dict(json_obj)

    @classmethod
    def from_yaml(cls, file_path):
        """
        returns a BaseObj initialised with yaml content
        from file_path
        """
        with open(file_path) as fp:
            yaml_obj = yaml.safe_load(fp)
            return cls.from_dict(yaml_obj)

    @classmethod
    def from_dict(cls, dict_value):
        """
        returns a BaseObj initialise with dict_value
        """
        bobj = cls()
        bobj.root = Dict(dict_value)
        return bobj

    def update_root(self, file_path):
        """
        update self.root with YAML/JSON content in file_path
        raises CompileError if file_path does not end with .yaml, .yml or .json
        """
        with open(file_path) as fp:
            if file_path.endswith(".yaml") or file_path.endswith(".yml"):
                yaml_obj = yaml.safe_load(fp)
                _copy = dict(self.root)
                _copy.update(yaml_obj)
                self.root = Dict(_copy)

            elif file_path.endswith(".json"):
                json_obj = json.load(fp)
                _copy = dict(self.root)
                _copy.update(json_obj)
                self.root = Dict(_copy)
            else:
                raise CompileError("file_path is neither JSON or YAML: {}".format(file_path))

    def need(self, key, msg="key and value needed"):
        """
        requires that key is set in self.kwargs
        errors with msg if key not set
        """
        err_msg = '{}: "{}": {}'.format(self.__class__.__name__, key, msg)
        if key not in self.kwargs:
            raise CompileError(err_msg)

    def new(self):
        """
        initialise need()ed keys for
        a new BaseObj
        """
        pass

    def body(self):
        """
        set values/logic for self.root
        """
        pass

    def _to_dict(self, obj):
        """
        recursively update obj should it contain other
        BaseObj values
        """
        if isinstance(obj, BaseObj):
            for k, v in obj.root.items():
                obj.root[k] = self._to_dict(v)
            # BaseObj needs to return to_dict()
            return obj.root.to_dict()
        elif isinstance(obj, list):
            obj = [self._to_dict(item) for item in obj]
            # list has no .to_dict, return itself
            return obj
        elif isinstance(obj, dict):
            for k, v in obj.items():
                obj[k] = self._to_dict(v)
            # dict has no .to_dict, return itself
            return obj

        # anything else, return itself
        return obj

    def to_dict(self):
        """
        returns object dict
        """
        return self._to_dict(self)



#!/usr/bin/env python3

# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"refs tests"

import base64
import os
import tempfile
import unittest

from kapitan.errors import RefError, RefFromFuncError, RefHashMismatchError
from kapitan.refs.base import PlainRef, RefController, RefParams, Revealer
from kapitan.refs.env import EnvRef, DEFAULT_ENV_REF_VAR_PREFIX
from kapitan.refs.base64 import Base64Ref
from kapitan.utils import get_entropy

REFS_HOME = tempfile.mkdtemp()
REF_CONTROLLER = RefController(REFS_HOME)
REVEALER = Revealer(REF_CONTROLLER)
REF_CONTROLLER_EMBEDDED = RefController(REFS_HOME, embed_refs=True)
REVEALER_EMBEDDED = Revealer(REF_CONTROLLER_EMBEDDED)


class Base64RefsTest(unittest.TestCase):
    "Test refs"

    def test_plain_ref_compile(self):
        "check plain ref compile() output is not hashed"
        tag = "?{plain:my/ref1_plain}"
        REF_CONTROLLER[tag] = PlainRef(b"ref plain data")
        ref_obj = REF_CONTROLLER[tag]
        compiled = ref_obj.compile()
        self.assertEqual(compiled, "ref plain data")

    def test_plain_ref_reveal(self):
        "check plain ref reveal() output is original plain data"
        tag = "?{plain:my/ref2_plain}"
        REF_CONTROLLER[tag] = PlainRef(b"ref 2 plain data")
        ref_obj = REF_CONTROLLER[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 2 plain data")

    def test_env_ref_compile(self):
        "check env ref compile() output is valid"
        tag = "?{env:my/ref1_env}"
        REF_CONTROLLER[tag] = EnvRef(b"ref 1 env data")
        ref_obj = REF_CONTROLLER[tag]
        revealed = ref_obj.compile()
        self.assertEqual(revealed, "?{env:my/ref1_env:877234e3}")

    def test_env_ref_reveal_default(self):
        "check env ref reveal() output is original plain env data when environment var is not set"
        tag = "?{env:my/ref1_env}"
        REF_CONTROLLER[tag] = EnvRef(b"ref 1 env data")
        ref_obj = REF_CONTROLLER[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 1 env data")

    def test_env_ref_reveal_from_env(self):
        "check env ref reveal() output is value of environment var data when environment var is set"
        tag = "?{env:my/ref1_env}"
        REF_CONTROLLER[tag] = EnvRef(b"ref 1 env data")
        ref_obj = REF_CONTROLLER[tag]
        os.environ["{}{}".format(DEFAULT_ENV_REF_VAR_PREFIX, "ref1_env")] = "ref 1 env data from EV"
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 1 env data from EV")

    def test_base64_ref_compile(self):
        "check ref compile() output is valid"
        tag = "?{base64:my/ref1}"
        REF_CONTROLLER[tag] = Base64Ref(b"ref 1 data")
        ref_obj = REF_CONTROLLER[tag]
        compiled = ref_obj.compile()
        self.assertEqual(compiled, "?{base64:my/ref1:3342a45c}")

    def test_base64_ref_embedded_compile(self):
        "check ref embedded compile() output is valid"
        tag = "?{base64:my/ref1}"
        REF_CONTROLLER_EMBEDDED[tag] = Base64Ref(b"ref 1 data")
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]
        compiled_embedded = ref_obj.compile()
        embedded_tag = (
            "?{base64:eyJkYXRhIjogImNtVm1JREVnWkdGMFlRPT0iLCAiZW5jb2"
            "RpbmciOiAib3JpZ2luYWwiLCAidHlwZSI6ICJiYXNlNjQifQ==:embedded}"
        )
        self.assertEqual(compiled_embedded, embedded_tag)

    def test_base64_ref_recompile(self):
        "check ref recompile() output is valid"
        tag = "?{base64:my/ref1}"
        ref_obj = REF_CONTROLLER[tag]
        compiled = ref_obj.compile()
        self.assertEqual(compiled, "?{base64:my/ref1:3342a45c}")

    def test_base64_ref_update_compile(self):
        "check ref update and compile() output is valid"
        tag = "?{base64:my/ref1}"
        REF_CONTROLLER[tag] = Base64Ref(b"ref 1 more data")
        ref_obj = REF_CONTROLLER[tag]
        compiled = ref_obj.compile()
        self.assertEqual(compiled, "?{base64:my/ref1:ed438a62}")

    def test_base64_ref_reveal(self):
        "check ref reveal() output is valid"
        tag = "?{base64:my/ref2}"
        REF_CONTROLLER[tag] = Base64Ref(b"ref 2 data")
        ref_obj = REF_CONTROLLER[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 2 data")

    def test_base64_ref_embedded_reveal(self):
        "check ref embedded reveal() output is valid"
        tag = "?{base64:my/ref2}"
        REF_CONTROLLER_EMBEDDED[tag] = Base64Ref(b"ref 2 data")
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 2 data")

    def test_base64_ref_embedded_reveal_encoding_original(self):
        "check ref embedded reveal() encoding metadata is persisted"
        tag = "?{base64:my/ref2}"
        REF_CONTROLLER_EMBEDDED[tag] = Base64Ref(b"ref 2 data")
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, "ref 2 data")
        self.assertEqual(ref_obj.encoding, "original")

    def test_base64_ref_embedded_reveal_encoding_base64(self):
        "check ref embedded reveal() encoding metadata is persisted"
        tag = "?{base64:my/ref3}"
        REF_CONTROLLER_EMBEDDED[tag] = Base64Ref(base64.b64encode(b"ref 3 data"), encoding="base64")
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]
        revealed = ref_obj.reveal()
        self.assertEqual(revealed, base64.b64encode(b"ref 3 data").decode())
        self.assertEqual(ref_obj.encoding, "base64")

    def test_base64_ref_non_existent_raises_KeyError(self):
        "check RefController raises KeyError for non existent Base64Ref"
        tag = "?{base64:non/existent}"
        with self.assertRaises(KeyError):
            REF_CONTROLLER[tag]

    def test_base64_ref_tag_type(self):
        "check ref tag type is Base64Ref"
        tag = "?{base64:my/ref3}"
        tag_type = REF_CONTROLLER.tag_type(tag)
        self.assertEqual(tag_type, Base64Ref)

    def test_base64_ref_tag_type_name(self):
        "check ref tag type name is ref"
        tag = "?{base64:my/ref4}"
        tag, token, func_str = REF_CONTROLLER.tag_params(tag)
        type_name = REF_CONTROLLER.token_type_name(token)
        self.assertEqual(tag, "?{base64:my/ref4}")
        self.assertEqual(token, "base64:my/ref4")
        self.assertEqual(func_str, None)
        self.assertEqual(type_name, "base64")

    def test_base64_ref_tag_func_name(self):
        "check ref tag func name is correct"
        tag = "?{base64:my/ref5||randomstr}"
        tag, token, func_str = REF_CONTROLLER.tag_params(tag)
        self.assertEqual(tag, "?{base64:my/ref5||randomstr}")
        self.assertEqual(token, "base64:my/ref5")
        self.assertEqual(func_str, "||randomstr")

    def test_base64_ref_embedded_attr(self):
        "check embedded ref has embed_refs set to True"
        tag = "?{base64:my/ref2}"
        REF_CONTROLLER_EMBEDDED[tag] = Base64Ref(b"ref 2 data")
        ref_obj = REF_CONTROLLER_EMBEDDED[tag]
        self.assertTrue(ref_obj.embed_refs)

    def test_base64_ref_attr(self):
        "check ref has embed_refs set to False"
        tag = "?{base64:my/ref2}"
        REF_CONTROLLER[tag] = Base64Ref(b"ref 2 data")
        ref_obj = REF_CONTROLLER[tag]
        self.assertFalse(ref_obj.embed_refs)

    def test_base64_ref_path(self):
        "check ref tag path is correct"
        tag = "?{base64:my/ref6}"
        tag, token, func_str = REF_CONTROLLER.tag_params(tag)
        self.assertEqual(tag, "?{base64:my/ref6}")
        self.assertEqual(token, "base64:my/ref6")
        self.assertEqual(func_str, None)
        REF_CONTROLLER[tag] = Base64Ref(b"ref 6 data")
        ref_obj = REF_CONTROLLER[tag]
        self.assertEqual(ref_obj.path, "my/ref6")

    def test_base64_ref_func_raise_RefFromFuncError(self):
        """
        check new ref tag with function raises RefFromFuncError
        and then creates it using RefParams()
        """
        tag = "?{base64:my/ref7||randomstr}"
        with self.assertRaises(RefFromFuncError):
            REF_CONTROLLER[tag]
        try:
            REF_CONTROLLER[tag]
        except RefFromFuncError:
            REF_CONTROLLER[tag] = RefParams()
        REF_CONTROLLER[tag]

    def test_base64_ref_revealer_reveal_raw_data_tag(self):
        "check Revealer reveals raw data"
        tag = "?{base64:my/ref2}"
        data = "data with {}, period.".format(tag)
        revealed_data = REVEALER.reveal_raw(data)
        self.assertEqual(revealed_data, "data with ref 2 data, period.")

    def test_base64_ref_revealer_reveal_raw_data_tag_compiled_hash(self):
        "check Revealer reveals raw data with compiled tag (with hash)"
        tag = "?{base64:my/ref2}"
        tag_compiled = REF_CONTROLLER[tag].compile()
        data = "data with {}, period.".format(tag_compiled)
        revealed_data = REVEALER.reveal_raw(data)
        self.assertEqual(revealed_data, "data with ref 2 data, period.")

    def test_base64_ref_revealer_reveal_raw_data_tag_compiled_hash_mismatch(self):
        """
        check Revealer reveals raises RefHashMismatchError
        on mismatch compiled tag hashes
        """
        tag_compiled_hash_mismatch = "?{base64:my/ref2:deadbeef}"
        with self.assertRaises(RefHashMismatchError):
            data = "data with {}, period.".format(tag_compiled_hash_mismatch)
            REVEALER.reveal_raw(data)

    def test_compile_subvars(self):
        """
        test that refs with sub-variables compile properly,
        and refs with different sub-variables stored in the same file has the same hash
        """
        subvar_tag1 = "?{base64:ref/subvars@var1}"
        subvar_tag2 = "?{base64:ref/subvars@var2}"
        REF_CONTROLLER["?{base64:ref/subvars}"] = Base64Ref(b"ref 1 data")
        ref_obj1 = REF_CONTROLLER[subvar_tag1]
        ref_obj2 = REF_CONTROLLER[subvar_tag2]
        self.assertEqual(ref_obj1.compile(), "?{base64:ref/subvars@var1:4357a29b}")
        self.assertEqual(ref_obj2.compile(), "?{base64:ref/subvars@var2:4357a29b}")

    def test_compile_embedded_subvar_path(self):
        """
        test that embedded refs with sub-variables have
        valid embedded_subvar_path key and value
        """
        subvar_tag = "?{base64:ref/subvars}"
        subvar_tag_var1 = "?{base64:ref/subvars@var1}"
        subvar_tag_var2 = "?{base64:ref/subvars@var2}"
        subvar_tag_var3 = "?{base64:ref/subvars@var3.var4}"
        REF_CONTROLLER_EMBEDDED[subvar_tag] = Base64Ref(b"I am not yaml, just testing")
        REF_CONTROLLER_EMBEDDED[subvar_tag_var1] = Base64Ref(b"I am not yaml, just testing")
        REF_CONTROLLER_EMBEDDED[subvar_tag_var2] = Base64Ref(b"I am not yaml, just testing")
        ref_obj1 = REF_CONTROLLER_EMBEDDED[subvar_tag_var1]
        ref_obj2 = REF_CONTROLLER_EMBEDDED[subvar_tag_var2]
        ref_obj3 = REF_CONTROLLER_EMBEDDED[subvar_tag_var3]

        self.assertTrue(ref_obj1.embed_refs)
        self.assertTrue(ref_obj2.embed_refs)
        self.assertTrue(ref_obj3.embed_refs)

        # now get compiled embedded ref tags from controller
        ref_obj1 = REF_CONTROLLER_EMBEDDED[ref_obj1.compile()]
        ref_obj2 = REF_CONTROLLER_EMBEDDED[ref_obj2.compile()]
        ref_obj3 = REF_CONTROLLER_EMBEDDED[ref_obj3.compile()]

        # and validate meta data
        self.assertEqual(ref_obj1.embedded_subvar_path, "var1")
        self.assertEqual(ref_obj2.embedded_subvar_path, "var2")
        self.assertEqual(ref_obj3.embedded_subvar_path, "var3.var4")

    def test_compile_embedded_subvars(self):
        """
        test that embedded refs with sub-variables compile properly,
        embedded refs with sub-variables will _not_ have equal compiled tags
        """
        subvar_tag = "?{base64:ref/subvars}"
        subvar_tag_var1 = "?{base64:ref/subvars@var1}"
        subvar_tag_var2 = "?{base64:ref/subvars@var2}"
        REF_CONTROLLER_EMBEDDED[subvar_tag] = Base64Ref(b"I am not yaml, just testing")
        REF_CONTROLLER_EMBEDDED[subvar_tag_var1] = Base64Ref(b"I am not yaml, just testing")
        REF_CONTROLLER_EMBEDDED[subvar_tag_var2] = Base64Ref(b"I am not yaml, just testing")
        ref_obj1 = REF_CONTROLLER_EMBEDDED[subvar_tag_var1]
        ref_obj2 = REF_CONTROLLER_EMBEDDED[subvar_tag_var2]
        self.assertNotEqual(ref_obj1.compile(), ref_obj2.compile())

    def test_reveal_subvars_raise_RefError(self):
        """
        test that reveal with sub-variable fails should the secret not
        be in valid yaml format
        """
        tag_to_save = "?{base64:ref/subvars_error}"
        yaml_secret = b"this is not yaml"
        REF_CONTROLLER[tag_to_save] = Base64Ref(yaml_secret)
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/subvars_error")))

        with self.assertRaises(RefError):
            tag_subvar = "?{base64:ref/subvars_error@var3.var4}"
            data = "message here: {}".format(tag_subvar)
            REVEALER.reveal_raw(data)

    def test_reveal_subvars(self):
        "write yaml secret, and access sub-variables in secrets"
        tag_to_save = "?{base64:ref/subvars}"
        yaml_secret = b"""
        var1:
          var2: hello
        var3:
          var4: world
        """
        REF_CONTROLLER[tag_to_save] = Base64Ref(yaml_secret)
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/subvars")))

        tag_subvar = "?{base64:ref/subvars@var1.var2}"
        data = "message here: {}".format(tag_subvar)
        revealed_data = REVEALER.reveal_raw(data)
        self.assertEqual("message here: hello", revealed_data)

        tag_subvar = "?{base64:ref/subvars@var3.var4}"
        data = "message here: {}".format(tag_subvar)
        revealed_data = REVEALER.reveal_raw(data)
        self.assertEqual("message here: world", revealed_data)

        with self.assertRaises(RefError):
            tag_subvar = "?{base64:ref/subvars@var3.varDoesntExist}"
            data = "message here: {}".format(tag_subvar)
            revealed_data = REVEALER.reveal_raw(data)

    def test_reveal_embedded_subvars(self):
        "write yaml ref data, and access sub-variables in embedded compiled refs"
        tag_to_save = "?{base64:ref/subvars}"
        tag_var1 = "?{base64:ref/subvars@var1.var2}"
        tag_var2 = "?{base64:ref/subvars@var3.var4}"
        tag_var_doesnt_exist = "?{base64:ref/subvars@var3.varDoesntExist}"
        yaml_secret = b"""
        var1:
          var2: hello
        var3:
          var4: world
        """
        REF_CONTROLLER_EMBEDDED[tag_to_save] = Base64Ref(yaml_secret)
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/subvars")))

        REF_CONTROLLER_EMBEDDED[tag_var1] = Base64Ref(yaml_secret)
        REF_CONTROLLER_EMBEDDED[tag_var2] = Base64Ref(yaml_secret)
        REF_CONTROLLER_EMBEDDED[tag_var_doesnt_exist] = Base64Ref(yaml_secret)

        ref_var1 = REF_CONTROLLER_EMBEDDED[tag_var1]
        ref_var2 = REF_CONTROLLER_EMBEDDED[tag_var2]
        ref_var_doesnt_exist = REF_CONTROLLER_EMBEDDED[tag_var_doesnt_exist]

        data = "message here: {}".format(ref_var1.compile())
        revealed_data = REVEALER_EMBEDDED.reveal_raw(data)
        self.assertEqual("message here: hello", revealed_data)

        data = "message here: {}".format(ref_var2.compile())
        revealed_data = REVEALER_EMBEDDED.reveal_raw(data)
        self.assertEqual("message here: world", revealed_data)

        with self.assertRaises(RefError):
            data = "message here: {}".format(ref_var_doesnt_exist.compile())
            revealed_data = REVEALER_EMBEDDED.reveal_raw(data)

    def test_ref_function_randomstr(self):
        "write randomstr to secret, confirm ref file exists, reveal and check"

        tag = "?{base64:ref/randomstr||randomstr}"
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/base64")))

        file_with_tags = tempfile.mktemp()
        with open(file_with_tags, "w") as fp:
            fp.write("?{base64:ref/randomstr}")
        revealed = REVEALER.reveal_raw_file(file_with_tags)
        self.assertEqual(len(revealed), 43)  # default length of token_urlsafe() string is 43
        self.assertTrue(get_entropy(revealed) > 4)

        # Test with parameter nbytes=16, correlating with string length 16
        tag = "?{base64:ref/randomstr||randomstr:16}"
        REF_CONTROLLER[tag] = RefParams()
        REVEALER._reveal_tag_without_subvar.cache_clear()
        revealed = REVEALER.reveal_raw_file(file_with_tags)
        self.assertEqual(len(revealed), 16)

    def test_ref_function_base64(self):
        "write randomstr to ref and base64, confirm ref file exists, reveal and check"

        tag = "?{base64:ref/base64||randomstr|base64}"
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/base64")))

        file_with_tags = tempfile.mktemp()
        with open(file_with_tags, "w") as fp:
            fp.write("?{base64:ref/base64}")
        revealed = REVEALER.reveal_raw_file(file_with_tags)
        # If the following succeeds, we guarantee that ref is base64-encoded
        self.assertEqual(base64.b64encode(base64.b64decode(revealed)).decode("UTF-8"), revealed)

    def test_ref_function_sha256(self):
        "write randomstr to ref and sha256, confirm ref file exists, reveal and check"

        tag = "?{base64:ref/sha256||randomstr|sha256}"
        REF_CONTROLLER[tag] = RefParams()
        self.assertTrue(os.path.isfile(os.path.join(REFS_HOME, "ref/sha256")))

        file_with_tags = tempfile.mktemp()
        with open(file_with_tags, "w") as fp:
            fp.write("?{base64:ref/sha256}")
        revealed = REVEALER.reveal_raw_file(file_with_tags)
        self.assertEqual(len(revealed), 64)
        try:
            int(revealed, 16)  # sha256 should convert to hex
        except ValueError:
            raise Exception("ref is not sha256 hash")

    # TODO write tests for RefController errors (lookups, etc..)



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"reference functions"

import base64
import hashlib
import logging
import secrets  # python secrets module

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519, rsa
from kapitan.errors import RefError

logger = logging.getLogger(__name__)


def eval_func(func_name, ctx, *func_params):

    func_lookup = {
        "randomstr": randomstr,
        "sha256": sha256,
        "ed25519": ed25519_private_key,
        "rsa": rsa_private_key,
        "rsapublic": rsa_public_key,
        "publickey": public_key,
        "reveal": reveal,
    }

    return func_lookup[func_name](ctx, *func_params)


def randomstr(ctx, nbytes=""):
    """
    generates a URL-safe text string, containing nbytes random bytes
    sets it to ctx.data
    """
    if nbytes:
        nbytes = int(nbytes)
        # Generate twice the amount of bytes asked for
        # and then trim the string to nbytes length if it's longer
        ctx.data = secrets.token_urlsafe(2 * nbytes)[:nbytes]

    else:
        ctx.data = secrets.token_urlsafe()


def sha256(ctx, salt=""):
    """sets ctx.data to salted sha256 hexdigest for input_value"""
    if ctx.data:
        salted_input_value = salt + ":" + ctx.data
        ctx.data = hashlib.sha256(salted_input_value.encode()).hexdigest()
    else:
        raise RefError(
            "Ref error: eval_func: nothing to sha256 hash; try " "something like '|randomstr|sha256'"
        )


def ed25519_private_key(ctx):
    """sets ctx.data to a ed25519 private key"""

    key = ed25519.Ed25519PrivateKey.generate()

    ctx.data = str(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        "utf-8",
    )


def rsa_private_key(ctx, key_size="4096"):
    """sets ctx.data to a RSA private key of key_size, default 4096"""
    rsa_key_size = int(key_size)

    key = rsa.generate_private_key(public_exponent=65537, key_size=rsa_key_size, backend=default_backend())

    ctx.data = str(
        key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ),
        "utf-8",
    )


def rsa_public_key(ctx):
    """Derives RSA public key from revealed private key"""
    if not ctx.data:
        raise RefError(
            "Ref error: eval_func: RSA public key cannot be derived; try "
            "something like '|reveal:path/to/encrypted_private_key|rsapublic'"
        )

    public_key(ctx)


def public_key(ctx):
    """Derives RSA public key from revealed private key"""
    if not ctx.data:
        raise RefError(
            "Ref error: eval_func: public key cannot be derived; try "
            "something like '|reveal:path/to/encrypted_private_key|publickey'"
        )

    data_dec = ctx.data
    if ctx.ref_encoding == "base64":
        data_dec = base64.b64decode(data_dec).decode()

    private_key = serialization.load_pem_private_key(
        data_dec.encode(), password=None, backend=default_backend()
    )
    public_key = private_key.public_key()

    ctx.data = str(
        public_key.public_bytes(
            encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
        ),
        "UTF-8",
    )


def reveal(ctx, secret_path):
    """
    decrypt and return data from secret_path
    """
    token_type_name = ctx.ref_controller.token_type_name(ctx.token)
    secret_tag = "?{{{}:{}}}".format(token_type_name, secret_path)
    try:
        ref_obj = ctx.ref_controller[secret_tag]
        ctx.ref_encoding = ref_obj.encoding
        ctx.data = ref_obj.reveal()
    except KeyError:
        raise RefError(
            f"|reveal function error: {secret_path} file in {ctx.token}|reveal:{secret_path} does not exist"
        )



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass
#

from reclass.values import item
from reclass.utils.dictpath import DictPath
from reclass.errors import ResolveError


class RefItem(item.ItemWithReferences):

    type = item.ItemTypes.REFERENCE

    def assembleRefs(self, context={}):
        super(RefItem, self).assembleRefs(context)
        try:
            self._refs.append(self._flatten_contents(context))
        except ResolveError as e:
            self.allRefs = False

    def _flatten_contents(self, context, inventory=None):
        result = [str(i.render(context, inventory)) for i in self.contents]
        return "".join(result)

    def _resolve(self, ref, context):
        path = DictPath(self._settings.delimiter, ref)
        try:
            return path.get_value(context)
        except (KeyError, TypeError) as e:
            raise ResolveError(ref)

    def render(self, context, inventory):
        #strings = [str(i.render(context, inventory)) for i in self.contents]
        #return self._resolve("".join(strings), context)
        return self._resolve(self._flatten_contents(context, inventory),
                             context)

    def __str__(self):
        strings = [str(i) for i in self.contents]
        rs = self._settings.reference_sentinels
        return '{0}{1}{2}'.format(rs[0], ''.join(strings), rs[1])



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass.utils.parameterdict import ParameterDict
from reclass.utils.parameterlist import ParameterList
from reclass.settings import Settings
from reclass.datatypes import Exports, Parameters
from reclass.errors import ParseError
import unittest

SETTINGS = Settings()

class TestInvQuery(unittest.TestCase):

    def test_overwrite_method(self):
        e = Exports({'alpha': { 'one': 1, 'two': 2}}, SETTINGS, '')
        d = {'alpha': { 'three': 3, 'four': 4}}
        e.overwrite(d)
        e.interpolate()
        self.assertEqual(e.as_dict(), d)

    def test_interpolate_types(self):
        e = Exports({'alpha': { 'one': 1, 'two': 2}, 'beta': [ 1, 2 ]}, SETTINGS, '')
        r = {'alpha': { 'one': 1, 'two': 2}, 'beta': [ 1, 2 ]}
        self.assertIs(type(e.as_dict()['alpha']), ParameterDict)
        self.assertIs(type(e.as_dict()['beta']), ParameterList)
        e.interpolate()
        self.assertIs(type(e.as_dict()['alpha']), dict)
        self.assertIs(type(e.as_dict()['beta']), list)
        self.assertEqual(e.as_dict(), r)

    def test_malformed_invquery(self):
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a exports:b == self:test_value ]'}, SETTINGS, '')
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a if exports:b self:test_value ]'}, SETTINGS, '')
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a if exports:b == ]'}, SETTINGS, '')
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a if exports:b == self:test_value and exports:c = self:test_value2 ]'}, SETTINGS, '')
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a if exports:b == self:test_value or exports:c == ]'}, SETTINGS, '')
        with self.assertRaises(ParseError):
            p = Parameters({'exp': '$[ exports:a if exports:b == self:test_value anddd exports:c == self:test_value2 ]'}, SETTINGS, '')

    def test_value_expr_invquery(self):
        e = {'node1': {'a': 1, 'b': 2}, 'node2': {'a': 3, 'b': 4}}
        p = Parameters({'exp': '$[ exports:a ]'}, SETTINGS, '')
        r = {'exp': {'node1': 1, 'node2': 3}}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_if_expr_invquery(self):
        e = {'node1': {'a': 1, 'b': 2}, 'node2': {'a': 3, 'b': 4}}
        p = Parameters({'exp': '$[ exports:a if exports:b == 4 ]'}, SETTINGS, '')
        r = {'exp': {'node2': 3}}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_if_expr_invquery_with_refs(self):
        e = {'node1': {'a': 1, 'b': 2}, 'node2': {'a': 3, 'b': 4}}
        p = Parameters({'exp': '$[ exports:a if exports:b == self:test_value ]', 'test_value': 2}, SETTINGS, '')
        r = {'exp': {'node1': 1}, 'test_value': 2}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_list_if_expr_invquery(self):
        e = {'node1': {'a': 1, 'b': 2}, 'node2': {'a': 3, 'b': 3}, 'node3': {'a': 3, 'b': 2}}
        p = Parameters({'exp': '$[ if exports:b == 2 ]'}, SETTINGS, '')
        r1 = {'exp': ['node1', 'node3']}
        r2 = {'exp': ['node3', 'node1']}
        p.interpolate(e)
        self.assertIn(p.as_dict(), [ r1, r2 ])

    def test_if_expr_invquery_wth_and(self):
        e = {'node1': {'a': 1, 'b': 4, 'c': False}, 'node2': {'a': 3, 'b': 4, 'c': True}}
        p = Parameters({'exp': '$[ exports:a if exports:b == 4 and exports:c == True ]'}, SETTINGS, '')
        r = {'exp': {'node2': 3}}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_if_expr_invquery_wth_or(self):
        e = {'node1': {'a': 1, 'b': 4}, 'node2': {'a': 3, 'b': 3}}
        p = Parameters({'exp': '$[ exports:a if exports:b == 4 or exports:b == 3 ]'}, SETTINGS, '')
        r = {'exp': {'node1': 1, 'node2': 3}}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_list_if_expr_invquery_with_and(self):
        e = {'node1': {'a': 1, 'b': 2, 'c': 'green'}, 'node2': {'a': 3, 'b': 3}, 'node3': {'a': 3, 'b': 2, 'c': 'red'}}
        p = Parameters({'exp': '$[ if exports:b == 2 and exports:c == green ]'}, SETTINGS, '')
        r = {'exp': ['node1']}
        p.interpolate(e)
        self.assertEqual(p.as_dict(), r)

    def test_list_if_expr_invquery_with_and_missing(self):
        inventory = {'node1': {'a': 1, 'b': 2, 'c': 'green'},
                     'node2': {'a': 3, 'b': 3},
                     'node3': {'a': 3, 'b': 2}}
        mapping = {'exp': '$[ if exports:b == 2 and exports:c == green ]'}
        expected = {'exp': ['node1']}

        pars = Parameters(mapping, SETTINGS, '')
        pars.interpolate(inventory)

        self.assertEqual(pars.as_dict(), expected)

    def test_list_if_expr_invquery_with_and(self):
        e = {'node1': {'a': 1, 'b': 2}, 'node2': {'a': 3, 'b': 3}, 'node3': {'a': 3, 'b': 4}}
        p = Parameters({'exp': '$[ if exports:b == 2 or exports:b == 4 ]'}, SETTINGS, '')
        r1 = {'exp': ['node1', 'node3']}
        r2 = {'exp': ['node3', 'node1']}
        p.interpolate(e)
        self.assertIn(p.as_dict(), [ r1, r2 ])

    def test_merging_inv_queries(self):
        e = {'node1': {'a': 1}, 'node2': {'a': 1}, 'node3': {'a': 2}}
        p1 = Parameters({'exp': '$[ if exports:a == 1 ]'}, SETTINGS, '')
        p2 = Parameters({'exp': '$[ if exports:a == 2 ]'}, SETTINGS, '')
        r = { 'exp': [ 'node1', 'node2', 'node3' ] }
        p1.merge(p2)
        p1.interpolate(e)
        self.assertEqual(p1.as_dict(), r)

if __name__ == '__main__':
    unittest.main()



# Copyright 2019 The Kapitan Authors
# SPDX-FileCopyrightText: 2020 The Kapitan Authors <kapitan-admins@googlegroups.com>
#
# SPDX-License-Identifier: Apache-2.0

"kapitan error classes"


class KapitanError(Exception):
    """generic kapitan error"""

    pass


class CompileError(KapitanError):
    """compile error"""

    pass


class InventoryError(KapitanError):
    """inventory error"""

    pass


class SecretError(KapitanError):
    """secrets error"""

    pass


class RefError(KapitanError):
    """ref error"""

    pass


class RefBackendError(KapitanError):
    """ref backend error"""

    pass


class RefFromFuncError(KapitanError):
    """ref from func error"""

    pass


class RefHashMismatchError(KapitanError):
    """ref has mismatch error"""

    pass


class HelmBindingUnavailableError(KapitanError):
    """helm input is used when the binding is not available"""

    pass


class HelmTemplateError(KapitanError):
    pass


class GitSubdirNotFoundError(KapitanError):
    """git dependency subdir not found error"""

    pass


class RequestUnsuccessfulError(KapitanError):
    """request error"""

    pass


class KubernetesManifestValidationError(KapitanError):
    """kubernetes manifest schema validation error"""

    pass



#
# -*- coding: utf-8 -*-
#
# This file is part of reclass (http://github.com/madduck/reclass)
#
# Copyright © 2007–14 martin f. krafft <madduck@madduck.net>
# Released under the terms of the Artistic Licence 2.0
#
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from reclass import datatypes
import yaml
import os
from reclass.errors import NotFoundError

_SafeLoader = yaml.CSafeLoader if yaml.__with_libyaml__ else yaml.SafeLoader

class YamlData(object):

    @classmethod
    def from_file(cls, path):
        ''' Initialise yaml data from a local file '''
        abs_path = os.path.abspath(path)
        if not os.path.isfile(abs_path):
            raise NotFoundError('No such file: %s' % abs_path)
        if not os.access(abs_path, os.R_OK):
            raise NotFoundError('Cannot open: %s' % abs_path)
        y = cls('yaml_fs://{0}'.format(abs_path))
        with open(abs_path) as fp:
            data = yaml.load(fp, Loader=_SafeLoader)
            if data is not None:
                y._data = data
        return y

    @classmethod
    def from_string(cls, string, uri):
        ''' Initialise yaml data from a string '''
        y = cls(uri)
        data = yaml.load(string, Loader=_SafeLoader)
        if data is not None:
            y._data = data
        return y

    def __init__(self, uri):
        self._uri = uri
        self._data = dict()

    uri = property(lambda self: self._uri)

    def get_data(self):
        return self._data

    def set_absolute_names(self, name, names):
        new_names = []
        for n in names:
            if n[0] == '.':
                dots = self.count_dots(n)
                levels_up = (dots * (-1))
                parent = '.'.join(name.split('.')[0:levels_up])
                if parent == '':
                    n = n[dots:]
                else:
                    n = parent + n[dots - 1:]
            new_names.append(n)
        return new_names

    def yield_dots(self, value):
        try:
            idx = value.index('.')
        except ValueError:
            return
        if idx == 0:
            yield '.'
            for dot in self.yield_dots(value[1:]):
                yield dot

    def count_dots(self, value):
        return len(list(self.yield_dots(value)))

    def get_entity(self, name, pathname, settings):
        classes = self._data.get('classes')
        if classes is None:
            classes = []
        classes = self.set_absolute_names(name, classes)
        classes = datatypes.Classes(classes)

        applications = self._data.get('applications')
        if applications is None:
            applications = []
        applications = datatypes.Applications(applications)

        parameters = self._data.get('parameters')
        if parameters is None:
            parameters = {}
        parameters = datatypes.Parameters(parameters, settings, self._uri)

        exports = self._data.get('exports')
        if exports is None:
            exports = {}
        exports = datatypes.Exports(exports, settings, self._uri)

        env = self._data.get('environment', None)

        return datatypes.Entity(settings, classes=classes, applications=applications, parameters=parameters,
                                exports=exports, name=name, pathname=pathname, environment=env, uri=self.uri)

    def __str__(self):
        return '<{0} {1}, {2}>'.format(self.__class__.__name__, self._uri,
                                       self._data)

    def __repr__(self):
        return '<{0} {1}, {2}>'.format(self.__class__.__name__, self._uri,
                                       self._data.keys())
